""" Multicast DNS Service Discovery for Python, v0.14-wmcbrine
    Copyright 2003 Paul Scott-Murphy, 2014 William McBrine

    This module provides a framework for the use of DNS Service Discovery
    using IP multicast.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
    USA

    COPIED FROM https://raw.githubusercontent.com/jstasiak/python-zeroconf/master/zeroconf.py
    because 0.20 is not released yet.
"""

import enum
import errno
import logging
import re
import select
import socket
import struct
import sys
import threading
import time
from functools import reduce

import netifaces

__author__ = 'Paul Scott-Murphy, William McBrine'
__maintainer__ = 'Jakub Stasiak <jakub@stasiak.at>'
__version__ = '0.19.1'
__license__ = 'LGPL'


__all__ = [
    "__version__",
    "Zeroconf", "ServiceInfo", "ServiceBrowser",
    "Error", "InterfaceChoice", "ServiceStateChange",
]


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

if log.level == logging.NOTSET:
    log.setLevel(logging.WARN)

# Some timing constants

_UNREGISTER_TIME = 125
_CHECK_TIME = 175
_REGISTER_TIME = 225
_LISTENER_TIME = 200
_BROWSER_TIME = 500

# Some DNS constants

_MDNS_ADDR = '224.0.0.251'
_MDNS_PORT = 5353
_DNS_PORT = 53
_DNS_TTL = 60 * 60  # one hour default TTL

_MAX_MSG_TYPICAL = 1460  # unused
_MAX_MSG_ABSOLUTE = 8966

_FLAGS_QR_MASK = 0x8000  # query response mask
_FLAGS_QR_QUERY = 0x0000  # query
_FLAGS_QR_RESPONSE = 0x8000  # response

_FLAGS_AA = 0x0400  # Authoritative answer
_FLAGS_TC = 0x0200  # Truncated
_FLAGS_RD = 0x0100  # Recursion desired
_FLAGS_RA = 0x8000  # Recursion available

_FLAGS_Z = 0x0040  # Zero
_FLAGS_AD = 0x0020  # Authentic data
_FLAGS_CD = 0x0010  # Checking disabled

_CLASS_IN = 1
_CLASS_CS = 2
_CLASS_CH = 3
_CLASS_HS = 4
_CLASS_NONE = 254
_CLASS_ANY = 255
_CLASS_MASK = 0x7FFF
_CLASS_UNIQUE = 0x8000

_TYPE_A = 1
_TYPE_NS = 2
_TYPE_MD = 3
_TYPE_MF = 4
_TYPE_CNAME = 5
_TYPE_SOA = 6
_TYPE_MB = 7
_TYPE_MG = 8
_TYPE_MR = 9
_TYPE_NULL = 10
_TYPE_WKS = 11
_TYPE_PTR = 12
_TYPE_HINFO = 13
_TYPE_MINFO = 14
_TYPE_MX = 15
_TYPE_TXT = 16
_TYPE_AAAA = 28
_TYPE_SRV = 33
_TYPE_ANY = 255

# Mapping constants to names

_CLASSES = {_CLASS_IN: "in",
            _CLASS_CS: "cs",
            _CLASS_CH: "ch",
            _CLASS_HS: "hs",
            _CLASS_NONE: "none",
            _CLASS_ANY: "any"}

_TYPES = {_TYPE_A: "a",
          _TYPE_NS: "ns",
          _TYPE_MD: "md",
          _TYPE_MF: "mf",
          _TYPE_CNAME: "cname",
          _TYPE_SOA: "soa",
          _TYPE_MB: "mb",
          _TYPE_MG: "mg",
          _TYPE_MR: "mr",
          _TYPE_NULL: "null",
          _TYPE_WKS: "wks",
          _TYPE_PTR: "ptr",
          _TYPE_HINFO: "hinfo",
          _TYPE_MINFO: "minfo",
          _TYPE_MX: "mx",
          _TYPE_TXT: "txt",
          _TYPE_AAAA: "quada",
          _TYPE_SRV: "srv",
          _TYPE_ANY: "any"}

_HAS_A_TO_Z = re.compile(r'[A-Za-z]')
_HAS_ONLY_A_TO_Z_NUM_HYPHEN = re.compile(r'^[A-Za-z0-9\-]+$')
_HAS_ASCII_CONTROL_CHARS = re.compile(r'[\x00-\x1f\x7f]')

int2byte = struct.Struct(">B").pack


@enum.unique
class InterfaceChoice(enum.Enum):
    Default = 1
    All = 2


@enum.unique
class ServiceStateChange(enum.Enum):
    Added = 1
    Removed = 2


HOST_ONLY_NETWORK_MASK = '255.255.255.255'


# utility functions


def current_time_millis():
    """Current system time in milliseconds"""
    return time.time() * 1000


def service_type_name(type_):
    """
    Validate a fully qualified service name, instance or subtype. [rfc6763]

    Returns fully qualified service name.

    Domain names used by mDNS-SD take the following forms:

                   <sn> . <_tcp|_udp> . local.
      <Instance> . <sn> . <_tcp|_udp> . local.
      <sub>._sub . <sn> . <_tcp|_udp> . local.

    1) must end with 'local.'

      This is true because we are implementing mDNS and since the 'm' means
      multi-cast, the 'local.' domain is mandatory.

    2) local is preceded with either '_udp.' or '_tcp.'

    3) service name <sn> precedes <_tcp|_udp>

      The rules for Service Names [RFC6335] state that they may be no more
      than fifteen characters long (not counting the mandatory underscore),
      consisting of only letters, digits, and hyphens, must begin and end
      with a letter or digit, must not contain consecutive hyphens, and
      must contain at least one letter.

    The instance name <Instance> and sub type <sub> may be up to 63 bytes.

    The portion of the Service Instance Name is a user-
    friendly name consisting of arbitrary Net-Unicode text [RFC5198]. It
    MUST NOT contain ASCII control characters (byte values 0x00-0x1F and
    0x7F) [RFC20] but otherwise is allowed to contain any characters,
    without restriction, including spaces, uppercase, lowercase,
    punctuation -- including dots -- accented characters, non-Roman text,
    and anything else that may be represented using Net-Unicode.

    :param type_: Type, SubType or service name to validate
    :return: fully qualified service name (eg: _http._tcp.local.)
    """
    if not (type_.endswith('._tcp.local.') or type_.endswith('._udp.local.')):
        raise BadTypeInNameException(
            "Type '%s' must end with '._tcp.local.' or '._udp.local.'" %
            type_)

    remaining = type_[:-len('._tcp.local.')].split('.')
    name = remaining.pop()
    if not name:
        raise BadTypeInNameException("No Service name found")

    if len(remaining) == 1 and len(remaining[0]) == 0:
        raise BadTypeInNameException(
            "Type '%s' must not start with '.'" % type_)

    if name[0] != '_':
        raise BadTypeInNameException(
            "Service name (%s) must start with '_'" % name)

    # remove leading underscore
    name = name[1:]

    if len(name) > 15:
        raise BadTypeInNameException(
            "Service name (%s) must be <= 15 bytes" % name)

    if '--' in name:
        raise BadTypeInNameException(
            "Service name (%s) must not contain '--'" % name)

    if '-' in (name[0], name[-1]):
        raise BadTypeInNameException(
            "Service name (%s) may not start or end with '-'" % name)

    if not _HAS_A_TO_Z.search(name):
        raise BadTypeInNameException(
            "Service name (%s) must contain at least one letter (eg: 'A-Z')" %
            name)

    if not _HAS_ONLY_A_TO_Z_NUM_HYPHEN.search(name):
        raise BadTypeInNameException(
            "Service name (%s) must contain only these characters: "
            "A-Z, a-z, 0-9, hyphen ('-')" % name)

    if remaining and remaining[-1] == '_sub':
        remaining.pop()
        if len(remaining) == 0 or len(remaining[0]) == 0:
            raise BadTypeInNameException(
                "_sub requires a subtype name")

    if len(remaining) > 1:
        remaining = ['.'.join(remaining)]

    if remaining:
        length = len(remaining[0].encode('utf-8'))
        if length > 63:
            raise BadTypeInNameException("Too long: '%s'" % remaining[0])

        if _HAS_ASCII_CONTROL_CHARS.search(remaining[0]):
            raise BadTypeInNameException(
                "Ascii control character 0x00-0x1F and 0x7F illegal in '%s'" %
                remaining[0])

    return '_' + name + type_[-len('._tcp.local.'):]


# Exceptions


class Error(Exception):
    pass


class IncomingDecodeError(Error):
    pass


class NonUniqueNameException(Error):
    pass


class NamePartTooLongException(Error):
    pass


class AbstractMethodException(Error):
    pass


class BadTypeInNameException(Error):
    pass

# implementation classes


class QuietLogger:
    _seen_logs = {}

    @classmethod
    def log_exception_warning(cls, logger_data=None):
        exc_info = sys.exc_info()
        exc_str = str(exc_info[1])
        if exc_str not in cls._seen_logs:
            # log at warning level the first time this is seen
            cls._seen_logs[exc_str] = exc_info
            logger = log.warning
        else:
            logger = log.debug
        if logger_data is not None:
            logger(*logger_data)
        logger('Exception occurred:', exc_info=exc_info)

    @classmethod
    def log_warning_once(cls, *args):
        msg_str = args[0]
        if msg_str not in cls._seen_logs:
            cls._seen_logs[msg_str] = 0
            logger = log.warning
        else:
            logger = log.debug
        cls._seen_logs[msg_str] += 1
        logger(*args)


class DNSEntry:

    """A DNS entry"""

    def __init__(self, name, type_, class_):
        self.key = name.lower()
        self.name = name
        self.type = type_
        self.class_ = class_ & _CLASS_MASK
        self.unique = (class_ & _CLASS_UNIQUE) != 0

    def __eq__(self, other):
        """Equality test on name, type, and class"""
        return (isinstance(other, DNSEntry) and
                self.name == other.name and
                self.type == other.type and
                self.class_ == other.class_)

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    @staticmethod
    def get_class_(class_):
        """Class accessor"""
        return _CLASSES.get(class_, "?(%s)" % class_)

    @staticmethod
    def get_type(t):
        """Type accessor"""
        return _TYPES.get(t, "?(%s)" % t)

    def to_string(self, hdr, other):
        """String representation with additional information"""
        result = "%s[%s,%s" % (hdr, self.get_type(self.type),
                               self.get_class_(self.class_))
        if self.unique:
            result += "-unique,"
        else:
            result += ","
        result += self.name
        if other is not None:
            result += ",%s]" % other
        else:
            result += "]"
        return result


class DNSQuestion(DNSEntry):

    """A DNS question entry"""

    def __init__(self, name, type_, class_):
        DNSEntry.__init__(self, name, type_, class_)

    def answered_by(self, rec):
        """Returns true if the question is answered by the record"""
        return (self.class_ == rec.class_ and
                (self.type == rec.type or self.type == _TYPE_ANY) and
                self.name == rec.name)

    def __repr__(self):
        """String representation"""
        return DNSEntry.to_string(self, "question", None)


class DNSRecord(DNSEntry):

    """A DNS record - like a DNS entry, but has a TTL"""

    def __init__(self, name, type_, class_, ttl):
        DNSEntry.__init__(self, name, type_, class_)
        self.ttl = ttl
        self.created = current_time_millis()

    def __eq__(self, other):
        """Abstract method"""
        raise AbstractMethodException

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def suppressed_by(self, msg):
        """Returns true if any answer in a message can suffice for the
        information held in this record."""
        for record in msg.answers:
            if self.suppressed_by_answer(record):
                return True
        return False

    def suppressed_by_answer(self, other):
        """Returns true if another record has same name, type and class,
        and if its TTL is at least half of this record's."""
        return self == other and other.ttl > (self.ttl / 2)

    def get_expiration_time(self, percent):
        """Returns the time at which this record will have expired
        by a certain percentage."""
        return self.created + (percent * self.ttl * 10)

    def get_remaining_ttl(self, now):
        """Returns the remaining TTL in seconds."""
        return max(0, (self.get_expiration_time(100) - now) / 1000.0)

    def is_expired(self, now):
        """Returns true if this record has expired."""
        return self.get_expiration_time(100) <= now

    def is_stale(self, now):
        """Returns true if this record is at least half way expired."""
        return self.get_expiration_time(50) <= now

    def reset_ttl(self, other):
        """Sets this record's TTL and created time to that of
        another record."""
        self.created = other.created
        self.ttl = other.ttl

    def write(self, out):
        """Abstract method"""
        raise AbstractMethodException

    def to_string(self, other):
        """String representation with additional information"""
        arg = "%s/%s,%s" % (
            self.ttl, self.get_remaining_ttl(current_time_millis()), other)
        return DNSEntry.to_string(self, "record", arg)


class DNSAddress(DNSRecord):

    """A DNS address record"""

    def __init__(self, name, type_, class_, ttl, address):
        DNSRecord.__init__(self, name, type_, class_, ttl)
        self.address = address

    def write(self, out):
        """Used in constructing an outgoing packet"""
        out.write_string(self.address)

    def __eq__(self, other):
        """Tests equality on address"""
        return (isinstance(other, DNSAddress) and DNSEntry.__eq__(self, other) and
                self.address == other.address)

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation"""
        try:
            return str(socket.inet_ntoa(self.address))
        except Exception:  # TODO stop catching all Exceptions
            return str(self.address)


class DNSHinfo(DNSRecord):

    """A DNS host information record"""

    def __init__(self, name, type_, class_, ttl, cpu, os):
        DNSRecord.__init__(self, name, type_, class_, ttl)
        try:
            self.cpu = cpu.decode('utf-8')
        except AttributeError:
            self.cpu = cpu
        try:
            self.os = os.decode('utf-8')
        except AttributeError:
            self.os = os

    def write(self, out):
        """Used in constructing an outgoing packet"""
        out.write_character_string(self.cpu.encode('utf-8'))
        out.write_character_string(self.os.encode('utf-8'))

    def __eq__(self, other):
        """Tests equality on cpu and os"""
        return (isinstance(other, DNSHinfo) and DNSEntry.__eq__(self, other) and
                self.cpu == other.cpu and self.os == other.os)

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation"""
        return self.cpu + " " + self.os


class DNSPointer(DNSRecord):

    """A DNS pointer record"""

    def __init__(self, name, type_, class_, ttl, alias):
        DNSRecord.__init__(self, name, type_, class_, ttl)
        self.alias = alias

    def write(self, out):
        """Used in constructing an outgoing packet"""
        out.write_name(self.alias)

    def __eq__(self, other):
        """Tests equality on alias"""
        return (isinstance(other, DNSPointer) and DNSEntry.__eq__(self, other) and
                self.alias == other.alias)

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation"""
        return self.to_string(self.alias)


class DNSText(DNSRecord):

    """A DNS text record"""

    def __init__(self, name, type_, class_, ttl, text):
        assert isinstance(text, (bytes, type(None)))
        DNSRecord.__init__(self, name, type_, class_, ttl)
        self.text = text

    def write(self, out):
        """Used in constructing an outgoing packet"""
        out.write_string(self.text)

    def __eq__(self, other):
        """Tests equality on text"""
        return (isinstance(other, DNSText) and DNSEntry.__eq__(self, other) and
                self.text == other.text)

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation"""
        if len(self.text) > 10:
            return self.to_string(self.text[:7]) + "..."
        else:
            return self.to_string(self.text)


class DNSService(DNSRecord):

    """A DNS service record"""

    def __init__(self, name, type_, class_, ttl,
                 priority, weight, port, server):
        DNSRecord.__init__(self, name, type_, class_, ttl)
        self.priority = priority
        self.weight = weight
        self.port = port
        self.server = server

    def write(self, out):
        """Used in constructing an outgoing packet"""
        out.write_short(self.priority)
        out.write_short(self.weight)
        out.write_short(self.port)
        out.write_name(self.server)

    def __eq__(self, other):
        """Tests equality on priority, weight, port and server"""
        return (isinstance(other, DNSService) and
                DNSEntry.__eq__(self, other) and
                self.priority == other.priority and
                self.weight == other.weight and
                self.port == other.port and
                self.server == other.server)

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation"""
        return self.to_string("%s:%s" % (self.server, self.port))


class DNSIncoming(QuietLogger):

    """Object representation of an incoming DNS packet"""

    def __init__(self, data):
        """Constructor from string holding bytes of packet"""
        self.offset = 0
        self.data = data
        self.questions = []
        self.answers = []
        self.id = 0
        self.flags = 0
        self.num_questions = 0
        self.num_answers = 0
        self.num_authorities = 0
        self.num_additionals = 0
        self.valid = False

        try:
            self.read_header()
            self.read_questions()
            self.read_others()
            self.valid = True

        except (IndexError, struct.error, IncomingDecodeError):
            self.log_exception_warning((
                'Choked at offset %d while unpacking %r', self.offset, data))

    def unpack(self, format_):
        length = struct.calcsize(format_)
        info = struct.unpack(
            format_, self.data[self.offset:self.offset + length])
        self.offset += length
        return info

    def read_header(self):
        """Reads header portion of packet"""
        (self.id, self.flags, self.num_questions, self.num_answers,
         self.num_authorities, self.num_additionals) = self.unpack(b'!6H')

    def read_questions(self):
        """Reads questions section of packet"""
        for i in range(self.num_questions):
            name = self.read_name()
            type_, class_ = self.unpack(b'!HH')

            question = DNSQuestion(name, type_, class_)
            self.questions.append(question)

    # def read_int(self):
    #     """Reads an integer from the packet"""
    #     return self.unpack(b'!I')[0]

    def read_character_string(self):
        """Reads a character string from the packet"""
        length = self.data[self.offset]
        self.offset += 1
        return self.read_string(length)

    def read_string(self, length):
        """Reads a string of a given length from the packet"""
        info = self.data[self.offset:self.offset + length]
        self.offset += length
        return info

    def read_unsigned_short(self):
        """Reads an unsigned short from the packet"""
        return self.unpack(b'!H')[0]

    def read_others(self):
        """Reads the answers, authorities and additionals section of the
        packet"""
        n = self.num_answers + self.num_authorities + self.num_additionals
        for i in range(n):
            domain = self.read_name()
            type_, class_, ttl, length = self.unpack(b'!HHiH')

            rec = None
            if type_ == _TYPE_A:
                rec = DNSAddress(
                    domain, type_, class_, ttl, self.read_string(4))
            elif type_ == _TYPE_CNAME or type_ == _TYPE_PTR:
                rec = DNSPointer(
                    domain, type_, class_, ttl, self.read_name())
            elif type_ == _TYPE_TXT:
                rec = DNSText(
                    domain, type_, class_, ttl, self.read_string(length))
            elif type_ == _TYPE_SRV:
                rec = DNSService(
                    domain, type_, class_, ttl,
                    self.read_unsigned_short(), self.read_unsigned_short(),
                    self.read_unsigned_short(), self.read_name())
            elif type_ == _TYPE_HINFO:
                rec = DNSHinfo(
                    domain, type_, class_, ttl,
                    self.read_character_string(), self.read_character_string())
            elif type_ == _TYPE_AAAA:
                rec = DNSAddress(
                    domain, type_, class_, ttl, self.read_string(16))
            else:
                # Try to ignore types we don't know about
                # Skip the payload for the resource record so the next
                # records can be parsed correctly
                self.offset += length

            if rec is not None:
                self.answers.append(rec)

    def is_query(self):
        """Returns true if this is a query"""
        return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_QUERY

    def is_response(self):
        """Returns true if this is a response"""
        return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_RESPONSE

    def read_utf(self, offset, length):
        """Reads a UTF-8 string of a given length from the packet"""
        return str(self.data[offset:offset + length], 'utf-8', 'replace')

    def read_name(self):
        """Reads a domain name from the packet"""
        result = ''
        off = self.offset
        next_ = -1
        first = off

        while True:
            length = self.data[off]
            off += 1
            if length == 0:
                break
            t = length & 0xC0
            if t == 0x00:
                result = ''.join((result, self.read_utf(off, length) + '.'))
                off += length
            elif t == 0xC0:
                if next_ < 0:
                    next_ = off + 1
                off = ((length & 0x3F) << 8) | self.data[off]
                if off >= first:
                    raise IncomingDecodeError(
                        "Bad domain name (circular) at %s" % (off,))
                first = off
            else:
                raise IncomingDecodeError("Bad domain name at %s" % (off,))

        if next_ >= 0:
            self.offset = next_
        else:
            self.offset = off

        return result


class DNSOutgoing:

    """Object representation of an outgoing packet"""

    def __init__(self, flags, multicast=True):
        self.finished = False
        self.id = 0
        self.multicast = multicast
        self.flags = flags
        self.names = {}
        self.data = []
        self.size = 12
        self.state = self.State.init

        self.questions = []
        self.answers = []
        self.authorities = []
        self.additionals = []

    def __repr__(self):
        return '<DNSOutgoing:{%s}>' % ', '.join([
            'multicast=%s' % self.multicast,
            'flags=%s' % self.flags,
            'questions=%s' % self.questions,
            'answers=%s' % self.answers,
            'authorities=%s' % self.authorities,
            'additionals=%s' % self.additionals,
        ])

    class State(enum.Enum):
        init = 0
        finished = 1

    def add_question(self, record):
        """Adds a question"""
        self.questions.append(record)

    def add_answer(self, inp, record):
        """Adds an answer"""
        if not record.suppressed_by(inp):
            self.add_answer_at_time(record, 0)

    def add_answer_at_time(self, record, now):
        """Adds an answer if it does not expire by a certain time"""
        if record is not None:
            if now == 0 or not record.is_expired(now):
                self.answers.append((record, now))

    def add_authorative_answer(self, record):
        """Adds an authoritative answer"""
        self.authorities.append(record)

    def add_additional_answer(self, record):
        """ Adds an additional answer

        From: RFC 6763, DNS-Based Service Discovery, February 2013

        12.  DNS Additional Record Generation

           DNS has an efficiency feature whereby a DNS server may place
           additional records in the additional section of the DNS message.
           These additional records are records that the client did not
           explicitly request, but the server has reasonable grounds to expect
           that the client might request them shortly, so including them can
           save the client from having to issue additional queries.

           This section recommends which additional records SHOULD be generated
           to improve network efficiency, for both Unicast and Multicast DNS-SD
           responses.

        12.1.  PTR Records

           When including a DNS-SD Service Instance Enumeration or Selective
           Instance Enumeration (subtype) PTR record in a response packet, the
           server/responder SHOULD include the following additional records:

           o  The SRV record(s) named in the PTR rdata.
           o  The TXT record(s) named in the PTR rdata.
           o  All address records (type "A" and "AAAA") named in the SRV rdata.

        12.2.  SRV Records

           When including an SRV record in a response packet, the
           server/responder SHOULD include the following additional records:

           o  All address records (type "A" and "AAAA") named in the SRV rdata.

        """
        self.additionals.append(record)

    def pack(self, format_, value):
        self.data.append(struct.pack(format_, value))
        self.size += struct.calcsize(format_)

    def write_byte(self, value):
        """Writes a single byte to the packet"""
        self.pack(b'!c', int2byte(value))

    def insert_short(self, index, value):
        """Inserts an unsigned short in a certain position in the packet"""
        self.data.insert(index, struct.pack(b'!H', value))
        self.size += 2

    def write_short(self, value):
        """Writes an unsigned short to the packet"""
        self.pack(b'!H', value)

    def write_int(self, value):
        """Writes an unsigned integer to the packet"""
        self.pack(b'!I', int(value))

    def write_string(self, value):
        """Writes a string to the packet"""
        assert isinstance(value, bytes)
        self.data.append(value)
        self.size += len(value)

    def write_utf(self, s):
        """Writes a UTF-8 string of a given length to the packet"""
        utfstr = s.encode('utf-8')
        length = len(utfstr)
        if length > 64:
            raise NamePartTooLongException
        self.write_byte(length)
        self.write_string(utfstr)

    def write_character_string(self, value):
        assert isinstance(value, bytes)
        length = len(value)
        if length > 256:
            raise NamePartTooLongException
        self.write_byte(length)
        self.write_string(value)

    def write_name(self, name):
        """
        Write names to packet

        18.14. Name Compression

        When generating Multicast DNS messages, implementations SHOULD use
        name compression wherever possible to compress the names of resource
        records, by replacing some or all of the resource record name with a
        compact two-byte reference to an appearance of that data somewhere
        earlier in the message [RFC1035].
        """

        # split name into each label
        parts = name.split('.')
        if not parts[-1]:
            parts.pop()

        # construct each suffix
        name_suffices = ['.'.join(parts[i:]) for i in range(len(parts))]

        # look for an existing name or suffix
        for count, sub_name in enumerate(name_suffices):
            if sub_name in self.names:
                break
        else:
            count += 1

        # note the new names we are saving into the packet
        for suffix in name_suffices[:count]:
            self.names[suffix] = self.size + len(name) - len(suffix) - 1

        # write the new names out.
        for part in parts[:count]:
            self.write_utf(part)

        # if we wrote part of the name, create a pointer to the rest
        if count != len(name_suffices):
            # Found substring in packet, create pointer
            index = self.names[name_suffices[count]]
            self.write_byte((index >> 8) | 0xC0)
            self.write_byte(index & 0xFF)
        else:
            # this is the end of a name
            self.write_byte(0)

    def write_question(self, question):
        """Writes a question to the packet"""
        self.write_name(question.name)
        self.write_short(question.type)
        self.write_short(question.class_)

    def write_record(self, record, now):
        """Writes a record (answer, authoritative answer, additional) to
        the packet"""
        if self.state == self.State.finished:
            return 1

        start_data_length, start_size = len(self.data), self.size
        self.write_name(record.name)
        self.write_short(record.type)
        if record.unique and self.multicast:
            self.write_short(record.class_ | _CLASS_UNIQUE)
        else:
            self.write_short(record.class_)
        if now == 0:
            self.write_int(record.ttl)
        else:
            self.write_int(record.get_remaining_ttl(now))
        index = len(self.data)

        # Adjust size for the short we will write before this record
        self.size += 2
        record.write(self)
        self.size -= 2

        length = sum((len(d) for d in self.data[index:]))
        # Here is the short we adjusted for
        self.insert_short(index, length)

        # if we go over, then rollback and quit
        if self.size > _MAX_MSG_ABSOLUTE:
            while len(self.data) > start_data_length:
                self.data.pop()
            self.size = start_size
            self.state = self.State.finished
            return 1
        return 0

    def packet(self):
        """Returns a string containing the packet's bytes

        No further parts should be added to the packet once this
        is done."""

        overrun_answers, overrun_authorities, overrun_additionals = 0, 0, 0

        if self.state != self.State.finished:
            for question in self.questions:
                self.write_question(question)
            for answer, time_ in self.answers:
                overrun_answers += self.write_record(answer, time_)
            for authority in self.authorities:
                overrun_authorities += self.write_record(authority, 0)
            for additional in self.additionals:
                overrun_additionals += self.write_record(additional, 0)
            self.state = self.State.finished

            self.insert_short(0, len(self.additionals) - overrun_additionals)
            self.insert_short(0, len(self.authorities) - overrun_authorities)
            self.insert_short(0, len(self.answers) - overrun_answers)
            self.insert_short(0, len(self.questions))
            self.insert_short(0, self.flags)
            if self.multicast:
                self.insert_short(0, 0)
            else:
                self.insert_short(0, self.id)
        return b''.join(self.data)


class DNSCache:

    """A cache of DNS entries"""

    def __init__(self):
        self.cache = {}

    def add(self, entry):
        """Adds an entry"""
        # Insert first in list so get returns newest entry
        self.cache.setdefault(entry.key, []).insert(0, entry)

    def remove(self, entry):
        """Removes an entry"""
        try:
            list_ = self.cache[entry.key]
            list_.remove(entry)
        except (KeyError, ValueError):
            pass

    def get(self, entry):
        """Gets an entry by key.  Will return None if there is no
        matching entry."""
        try:
            list_ = self.cache[entry.key]
            for cached_entry in list_:
                if entry.__eq__(cached_entry):
                    return cached_entry
        except (KeyError, ValueError):
            return None

    def get_by_details(self, name, type_, class_):
        """Gets an entry by details.  Will return None if there is
        no matching entry."""
        entry = DNSEntry(name, type_, class_)
        return self.get(entry)

    def entries_with_name(self, name):
        """Returns a list of entries whose key matches the name."""
        try:
            return self.cache[name.lower()]
        except KeyError:
            return []

    def current_entry_with_name_and_alias(self, name, alias):
        now = current_time_millis()
        for record in self.entries_with_name(name):
            if (record.type == _TYPE_PTR and
                    not record.is_expired(now) and
                    record.alias == alias):
                return record

    def entries(self):
        """Returns a list of all entries"""
        if not self.cache:
            return []
        else:
            # avoid size change during iteration by copying the cache
            values = list(self.cache.values())
            return reduce(lambda a, b: a + b, values)


class Engine(threading.Thread):

    """An engine wraps read access to sockets, allowing objects that
    need to receive data from sockets to be called back when the
    sockets are ready.

    A reader needs a handle_read() method, which is called when the socket
    it is interested in is ready for reading.

    Writers are not implemented here, because we only send short
    packets.
    """

    def __init__(self, zc):
        threading.Thread.__init__(self, name='zeroconf-Engine')
        self.daemon = True
        self.zc = zc
        self.readers = {}  # maps socket to reader
        self.timeout = 5
        self.condition = threading.Condition()
        self.start()

    def run(self):
        while not self.zc.done:
            with self.condition:
                rs = self.readers.keys()
                if len(rs) == 0:
                    # No sockets to manage, but we wait for the timeout
                    # or addition of a socket
                    self.condition.wait(self.timeout)

            if len(rs) != 0:
                try:
                    rr, wr, er = select.select(rs, [], [], self.timeout)
                    if not self.zc.done:
                        for socket_ in rr:
                            reader = self.readers.get(socket_)
                            if reader:
                                reader.handle_read(socket_)

                except (select.error, socket.error) as e:
                    # If the socket was closed by another thread, during
                    # shutdown, ignore it and exit
                    if e.args[0] != socket.EBADF or not self.zc.done:
                        raise

    def add_reader(self, reader, socket_):
        with self.condition:
            self.readers[socket_] = reader
            self.condition.notify()

    def del_reader(self, socket_):
        with self.condition:
            del self.readers[socket_]
            self.condition.notify()


class Listener(QuietLogger):

    """A Listener is used by this module to listen on the multicast
    group to which DNS messages are sent, allowing the implementation
    to cache information as it arrives.

    It requires registration with an Engine object in order to have
    the read() method called when a socket is available for reading."""

    def __init__(self, zc):
        self.zc = zc
        self.data = None

    def handle_read(self, socket_):
        try:
            data, (addr, port) = socket_.recvfrom(_MAX_MSG_ABSOLUTE)
        except Exception:
            self.log_exception_warning()
            return

        log.debug('Received from %r:%r: %r ', addr, port, data)

        self.data = data
        msg = DNSIncoming(data)
        if not msg.valid:
            pass

        elif msg.is_query():
            # Always multicast responses
            if port == _MDNS_PORT:
                self.zc.handle_query(msg, _MDNS_ADDR, _MDNS_PORT)

            # If it's not a multicast query, reply via unicast
            # and multicast
            elif port == _DNS_PORT:
                self.zc.handle_query(msg, addr, port)
                self.zc.handle_query(msg, _MDNS_ADDR, _MDNS_PORT)

        else:
            self.zc.handle_response(msg)


class Reaper(threading.Thread):

    """A Reaper is used by this module to remove cache entries that
    have expired."""

    def __init__(self, zc):
        threading.Thread.__init__(self, name='zeroconf-Reaper')
        self.daemon = True
        self.zc = zc
        self.start()

    def run(self):
        while True:
            self.zc.wait(10 * 1000)
            if self.zc.done:
                return
            now = current_time_millis()
            for record in self.zc.cache.entries():
                if record.is_expired(now):
                    self.zc.update_record(now, record)
                    self.zc.cache.remove(record)


class Signal:
    def __init__(self):
        self._handlers = []

    def fire(self, **kwargs):
        for h in list(self._handlers):
            h(**kwargs)

    @property
    def registration_interface(self):
        return SignalRegistrationInterface(self._handlers)


class SignalRegistrationInterface:

    def __init__(self, handlers):
        self._handlers = handlers

    def register_handler(self, handler):
        self._handlers.append(handler)
        return self

    def unregister_handler(self, handler):
        self._handlers.remove(handler)
        return self


class ServiceBrowser(threading.Thread):

    """Used to browse for a service of a specific type.

    The listener object will have its add_service() and
    remove_service() methods called when this browser
    discovers changes in the services availability."""

    def __init__(self, zc, type_, handlers=None, listener=None):
        """Creates a browser for a specific type"""
        assert handlers or listener, 'You need to specify at least one handler'
        if not type_.endswith(service_type_name(type_)):
            raise BadTypeInNameException
        threading.Thread.__init__(
            self, name='zeroconf-ServiceBrowser_' + type_)
        self.daemon = True
        self.zc = zc
        self.type = type_
        self.services = {}
        self.next_time = current_time_millis()
        self.delay = _BROWSER_TIME
        self._handlers_to_call = []

        self._service_state_changed = Signal()

        self.done = False

        if hasattr(handlers, 'add_service'):
            listener = handlers
            handlers = None

        handlers = handlers or []

        if listener:
            def on_change(zeroconf, service_type, name, state_change):
                args = (zeroconf, service_type, name)
                if state_change is ServiceStateChange.Added:
                    listener.add_service(*args)
                elif state_change is ServiceStateChange.Removed:
                    listener.remove_service(*args)
                else:
                    raise NotImplementedError(state_change)
            handlers.append(on_change)

        for h in handlers:
            self.service_state_changed.register_handler(h)

        self.start()

    @property
    def service_state_changed(self):
        return self._service_state_changed.registration_interface

    def update_record(self, zc, now, record):
        """Callback invoked by Zeroconf when new information arrives.

        Updates information required by browser in the Zeroconf cache."""

        def enqueue_callback(state_change, name):
            self._handlers_to_call.append(
                lambda zeroconf: self._service_state_changed.fire(
                    zeroconf=zeroconf,
                    service_type=self.type,
                    name=name,
                    state_change=state_change,
                ))

        if record.type == _TYPE_PTR and record.name == self.type:
            expired = record.is_expired(now)
            service_key = record.alias.lower()
            try:
                old_record = self.services[service_key]
            except KeyError:
                if not expired:
                    self.services[service_key] = record
                    enqueue_callback(ServiceStateChange.Added, record.alias)
            else:
                if not expired:
                    old_record.reset_ttl(record)
                else:
                    del self.services[service_key]
                    enqueue_callback(ServiceStateChange.Removed, record.alias)
                    return

            expires = record.get_expiration_time(75)
            if expires < self.next_time:
                self.next_time = expires

    def cancel(self):
        self.done = True
        self.zc.remove_listener(self)
        self.join()

    def run(self):
        self.zc.add_listener(self, DNSQuestion(self.type, _TYPE_PTR, _CLASS_IN))

        while True:
            now = current_time_millis()
            if len(self._handlers_to_call) == 0 and self.next_time > now:
                self.zc.wait(self.next_time - now)
            if self.zc.done or self.done:
                return
            now = current_time_millis()
            if self.next_time <= now:
                out = DNSOutgoing(_FLAGS_QR_QUERY)
                out.add_question(DNSQuestion(self.type, _TYPE_PTR, _CLASS_IN))
                for record in self.services.values():
                    if not record.is_stale(now):
                        out.add_answer_at_time(record, now)

                self.zc.send(out)
                self.next_time = now + self.delay
                self.delay = min(20 * 1000, self.delay * 2)

            if len(self._handlers_to_call) > 0 and not self.zc.done:
                handler = self._handlers_to_call.pop(0)
                handler(self.zc)


class ServiceInfo:

    """Service information"""

    def __init__(self, type_, name, address=None, port=None, weight=0,
                 priority=0, properties=None, server=None):
        """Create a service description.

        type_: fully qualified service type name
        name: fully qualified service name
        address: IP address as unsigned short, network byte order
        port: port that the service runs on
        weight: weight of the service
        priority: priority of the service
        properties: dictionary of properties (or a string holding the
                    bytes for the text field)
        server: fully qualified name for service host (defaults to name)"""

        if not type_.endswith(service_type_name(name)):
            raise BadTypeInNameException
        self.type = type_
        self.name = name
        self.address = address
        self.port = port
        self.weight = weight
        self.priority = priority
        if server:
            self.server = server
        else:
            self.server = name
        self._properties = {}
        self._set_properties(properties)

    @property
    def properties(self):
        return self._properties

    def _set_properties(self, properties):
        """Sets properties and text of this info from a dictionary"""
        if isinstance(properties, dict):
            self._properties = properties
            list_ = []
            result = b''
            for key, value in properties.items():
                if isinstance(key, str):
                    key = key.encode('utf-8')

                if value is None:
                    suffix = b''
                elif isinstance(value, str):
                    suffix = value.encode('utf-8')
                elif isinstance(value, bytes):
                    suffix = value
                elif isinstance(value, int):
                    if value:
                        suffix = b'true'
                    else:
                        suffix = b'false'
                else:
                    suffix = b''
                list_.append(b'='.join((key, suffix)))
            for item in list_:
                result = b''.join((result, int2byte(len(item)), item))
            self.text = result
        else:
            self.text = properties

    def _set_text(self, text):
        """Sets properties and text given a text field"""
        self.text = text
        result = {}
        end = len(text)
        index = 0
        strs = []
        while index < end:
            length = text[index]
            index += 1
            strs.append(text[index:index + length])
            index += length

        for s in strs:
            parts = s.split(b'=', 1)
            try:
                key, value = parts
            except ValueError:
                # No equals sign at all
                key = s
                value = False
            else:
                if value == b'true':
                    value = True
                elif value == b'false' or not value:
                    value = False

            # Only update non-existent properties
            if key and result.get(key) is None:
                result[key] = value

        self._properties = result

    def get_name(self):
        """Name accessor"""
        if self.type is not None and self.name.endswith("." + self.type):
            return self.name[:len(self.name) - len(self.type) - 1]
        return self.name

    def update_record(self, zc, now, record):
        """Updates service information from a DNS record"""
        if record is not None and not record.is_expired(now):
            if record.type == _TYPE_A:
                # if record.name == self.name:
                if record.name == self.server:
                    self.address = record.address
            elif record.type == _TYPE_SRV:
                if record.name == self.name:
                    self.server = record.server
                    self.port = record.port
                    self.weight = record.weight
                    self.priority = record.priority
                    # self.address = None
                    self.update_record(
                        zc, now, zc.cache.get_by_details(
                            self.server, _TYPE_A, _CLASS_IN))
            elif record.type == _TYPE_TXT:
                if record.name == self.name:
                    self._set_text(record.text)

    def request(self, zc, timeout):
        """Returns true if the service could be discovered on the
        network, and updates this object with details discovered.
        """
        now = current_time_millis()
        delay = _LISTENER_TIME
        next_ = now + delay
        last = now + timeout

        record_types_for_check_cache = [
            (_TYPE_SRV, _CLASS_IN),
            (_TYPE_TXT, _CLASS_IN),
        ]
        if self.server is not None:
            record_types_for_check_cache.append((_TYPE_A, _CLASS_IN))
        for record_type in record_types_for_check_cache:
            cached = zc.cache.get_by_details(self.name, *record_type)
            if cached:
                self.update_record(zc, now, cached)

        if None not in (self.server, self.address, self.text):
            return True

        try:
            zc.add_listener(self, DNSQuestion(self.name, _TYPE_ANY, _CLASS_IN))
            while None in (self.server, self.address, self.text):
                if last <= now:
                    return False
                if next_ <= now:
                    out = DNSOutgoing(_FLAGS_QR_QUERY)
                    out.add_question(
                        DNSQuestion(self.name, _TYPE_SRV, _CLASS_IN))
                    out.add_answer_at_time(
                        zc.cache.get_by_details(
                            self.name, _TYPE_SRV, _CLASS_IN), now)

                    out.add_question(
                        DNSQuestion(self.name, _TYPE_TXT, _CLASS_IN))
                    out.add_answer_at_time(
                        zc.cache.get_by_details(
                            self.name, _TYPE_TXT, _CLASS_IN), now)

                    if self.server is not None:
                        out.add_question(
                            DNSQuestion(self.server, _TYPE_A, _CLASS_IN))
                        out.add_answer_at_time(
                            zc.cache.get_by_details(
                                self.server, _TYPE_A, _CLASS_IN), now)
                    zc.send(out)
                    next_ = now + delay
                    delay *= 2

                zc.wait(min(next_, last) - now)
                now = current_time_millis()
        finally:
            zc.remove_listener(self)

        return True

    def __eq__(self, other):
        """Tests equality of service name"""
        return isinstance(other, ServiceInfo) and other.name == self.name

    def __ne__(self, other):
        """Non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation"""
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join(
                '%s=%r' % (name, getattr(self, name))
                for name in (
                    'type', 'name', 'address', 'port', 'weight', 'priority',
                    'server', 'properties',
                )
            )
        )


class ZeroconfServiceTypes:
    """
    Return all of the advertised services on any local networks
    """
    def __init__(self):
        self.found_services = set()

    def add_service(self, zc, type_, name):
        self.found_services.add(name)

    def remove_service(self, zc, type_, name):
        pass

    @classmethod
    def find(cls, zc=None, timeout=5, interfaces=InterfaceChoice.All):
        """
        Return all of the advertised services on any local networks.

        :param zc: Zeroconf() instance.  Pass in if already have an
                instance running or if non-default interfaces are needed
        :param timeout: seconds to wait for any responses
        :return: tuple of service type strings
        """
        local_zc = zc or Zeroconf(interfaces=interfaces)
        listener = cls()
        browser = ServiceBrowser(
            local_zc, '_services._dns-sd._udp.local.', listener=listener)

        # wait for responses
        time.sleep(timeout)

        # close down anything we opened
        if zc is None:
            local_zc.close()
        else:
            browser.cancel()

        return tuple(sorted(listener.found_services))


def get_all_addresses(address_family):
    return list(set(
        addr['addr']
        for iface in netifaces.interfaces()
        for addr in netifaces.ifaddresses(iface).get(address_family, [])
        if addr.get('netmask') != HOST_ONLY_NETWORK_MASK
    ))


def normalize_interface_choice(choice, address_family):
    if choice is InterfaceChoice.Default:
        choice = ['0.0.0.0']
    elif choice is InterfaceChoice.All:
        choice = get_all_addresses(address_family)
    return choice


def new_socket():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # SO_REUSEADDR should be equivalent to SO_REUSEPORT for
    # multicast UDP sockets (p 731, "TCP/IP Illustrated,
    # Volume 2"), but some BSD-derived systems require
    # SO_REUSEPORT to be specified explicity.  Also, not all
    # versions of Python have SO_REUSEPORT available.
    # Catch OSError and socket.error for kernel versions <3.9 because lacking
    # SO_REUSEPORT support.
    try:
        reuseport = socket.SO_REUSEPORT
    except AttributeError:
        pass
    else:
        try:
            s.setsockopt(socket.SOL_SOCKET, reuseport, 1)
        except (OSError, socket.error) as err:
            # OSError on python 3, socket.error on python 2
            if not err.errno == errno.ENOPROTOOPT:
                raise

    # OpenBSD needs the ttl and loop values for the IP_MULTICAST_TTL and
    # IP_MULTICAST_LOOP socket options as an unsigned char.
    ttl = struct.pack(b'B', 255)
    s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
    loop = struct.pack(b'B', 1)
    s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, loop)

    s.bind(('', _MDNS_PORT))
    return s


def get_errno(e):
    assert isinstance(e, socket.error)
    return e.args[0]


class Zeroconf(QuietLogger):

    """Implementation of Zeroconf Multicast DNS Service Discovery

    Supports registration, unregistration, queries and browsing.
    """

    def __init__(
        self,
        interfaces=InterfaceChoice.All,
    ):
        """Creates an instance of the Zeroconf class, establishing
        multicast communications, listening and reaping threads.

        :type interfaces: :class:`InterfaceChoice` or sequence of ip addresses
        """
        # hook for threads
        self._GLOBAL_DONE = False

        self._listen_socket = new_socket()
        interfaces = normalize_interface_choice(interfaces, socket.AF_INET)

        self._respond_sockets = []

        for i in interfaces:
            log.debug('Adding %r to multicast group', i)
            try:
                self._listen_socket.setsockopt(
                    socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP,
                    socket.inet_aton(_MDNS_ADDR) + socket.inet_aton(i))
            except socket.error as e:
                if get_errno(e) == errno.EADDRINUSE:
                    log.info(
                        'Address in use when adding %s to multicast group, '
                        'it is expected to happen on some systems', i,
                    )
                elif get_errno(e) == errno.EADDRNOTAVAIL:
                    log.info(
                        'Address not available when adding %s to multicast '
                        'group, it is expected to happen on some systems', i,
                    )
                    continue
                else:
                    raise

            respond_socket = new_socket()
            respond_socket.setsockopt(
                socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(i))

            self._respond_sockets.append(respond_socket)

        self.listeners = []
        self.browsers = {}
        self.services = {}
        self.servicetypes = {}

        self.cache = DNSCache()

        self.condition = threading.Condition()

        self.engine = Engine(self)
        self.listener = Listener(self)
        self.engine.add_reader(self.listener, self._listen_socket)
        self.reaper = Reaper(self)

        self.debug = None

    @property
    def done(self):
        return self._GLOBAL_DONE

    def wait(self, timeout):
        """Calling thread waits for a given number of milliseconds or
        until notified."""
        with self.condition:
            self.condition.wait(timeout / 1000.0)

    def notify_all(self):
        """Notifies all waiting threads"""
        with self.condition:
            self.condition.notify_all()

    def get_service_info(self, type_, name, timeout=3000):
        """Returns network's service information for a particular
        name and type, or None if no service matches by the timeout,
        which defaults to 3 seconds."""
        info = ServiceInfo(type_, name)
        if info.request(self, timeout):
            return info

    def add_service_listener(self, type_, listener):
        """Adds a listener for a particular service type.  This object
        will then have its update_record method called when information
        arrives for that type."""
        self.remove_service_listener(listener)
        self.browsers[listener] = ServiceBrowser(self, type_, listener)

    def remove_service_listener(self, listener):
        """Removes a listener from the set that is currently listening."""
        if listener in self.browsers:
            self.browsers[listener].cancel()
            del self.browsers[listener]

    def remove_all_service_listeners(self):
        """Removes a listener from the set that is currently listening."""
        for listener in [k for k in self.browsers]:
            self.remove_service_listener(listener)

    def register_service(self, info, ttl=_DNS_TTL, allow_name_change=False):
        """Registers service information to the network with a default TTL
        of 60 seconds.  Zeroconf will then respond to requests for
        information for that service.  The name of the service may be
        changed if needed to make it unique on the network."""
        self.check_service(info, allow_name_change)
        self.services[info.name.lower()] = info
        if info.type in self.servicetypes:
            self.servicetypes[info.type] += 1
        else:
            self.servicetypes[info.type] = 1
        now = current_time_millis()
        next_time = now
        i = 0
        while i < 3:
            if now < next_time:
                self.wait(next_time - now)
                now = current_time_millis()
                continue
            out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
            out.add_answer_at_time(
                DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, ttl, info.name), 0)
            out.add_answer_at_time(
                DNSService(info.name, _TYPE_SRV, _CLASS_IN,
                           ttl, info.priority, info.weight, info.port,
                           info.server), 0)

            out.add_answer_at_time(
                DNSText(info.name, _TYPE_TXT, _CLASS_IN, ttl, info.text), 0)
            if info.address:
                out.add_answer_at_time(
                    DNSAddress(info.server, _TYPE_A, _CLASS_IN,
                               ttl, info.address), 0)
            self.send(out)
            i += 1
            next_time += _REGISTER_TIME

    def unregister_service(self, info):
        """Unregister a service."""
        try:
            del self.services[info.name.lower()]
            if self.servicetypes[info.type] > 1:
                self.servicetypes[info.type] -= 1
            else:
                del self.servicetypes[info.type]
        except Exception as e:  # TODO stop catching all Exceptions
            log.exception('Unknown error, possibly benign: %r', e)
        now = current_time_millis()
        next_time = now
        i = 0
        while i < 3:
            if now < next_time:
                self.wait(next_time - now)
                now = current_time_millis()
                continue
            out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
            out.add_answer_at_time(
                DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, 0, info.name), 0)
            out.add_answer_at_time(
                DNSService(info.name, _TYPE_SRV, _CLASS_IN, 0,
                           info.priority, info.weight, info.port, info.name), 0)
            out.add_answer_at_time(
                DNSText(info.name, _TYPE_TXT, _CLASS_IN, 0, info.text), 0)

            if info.address:
                out.add_answer_at_time(
                    DNSAddress(info.server, _TYPE_A, _CLASS_IN, 0,
                               info.address), 0)
            self.send(out)
            i += 1
            next_time += _UNREGISTER_TIME

    def unregister_all_services(self):
        """Unregister all registered services."""
        if len(self.services) > 0:
            now = current_time_millis()
            next_time = now
            i = 0
            while i < 3:
                if now < next_time:
                    self.wait(next_time - now)
                    now = current_time_millis()
                    continue
                out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
                for info in self.services.values():
                    out.add_answer_at_time(DNSPointer(
                        info.type, _TYPE_PTR, _CLASS_IN, 0, info.name), 0)
                    out.add_answer_at_time(DNSService(
                        info.name, _TYPE_SRV, _CLASS_IN, 0,
                        info.priority, info.weight, info.port, info.server), 0)
                    out.add_answer_at_time(DNSText(
                        info.name, _TYPE_TXT, _CLASS_IN, 0, info.text), 0)
                    if info.address:
                        out.add_answer_at_time(DNSAddress(
                            info.server, _TYPE_A, _CLASS_IN, 0,
                            info.address), 0)
                self.send(out)
                i += 1
                next_time += _UNREGISTER_TIME

    def check_service(self, info, allow_name_change):
        """Checks the network for a unique service name, modifying the
        ServiceInfo passed in if it is not unique."""

        # This is kind of funky because of the subtype based tests
        # need to make subtypes a first class citizen
        service_name = service_type_name(info.name)
        if not info.type.endswith(service_name):
            raise BadTypeInNameException

        instance_name = info.name[:-len(service_name) - 1]
        next_instance_number = 2

        now = current_time_millis()
        next_time = now
        i = 0
        while i < 3:
            # check for a name conflict
            while self.cache.current_entry_with_name_and_alias(
                    info.type, info.name):
                if not allow_name_change:
                    raise NonUniqueNameException

                # change the name and look for a conflict
                info.name = '%s-%s.%s' % (
                    instance_name, next_instance_number, info.type)
                next_instance_number += 1
                service_type_name(info.name)
                next_time = now
                i = 0

            if now < next_time:
                self.wait(next_time - now)
                now = current_time_millis()
                continue

            out = DNSOutgoing(_FLAGS_QR_QUERY | _FLAGS_AA)
            self.debug = out
            out.add_question(DNSQuestion(info.type, _TYPE_PTR, _CLASS_IN))
            out.add_authorative_answer(DNSPointer(
                info.type, _TYPE_PTR, _CLASS_IN, _DNS_TTL, info.name))
            self.send(out)
            i += 1
            next_time += _CHECK_TIME

    def add_listener(self, listener, question):
        """Adds a listener for a given question.  The listener will have
        its update_record method called when information is available to
        answer the question."""
        now = current_time_millis()
        self.listeners.append(listener)
        if question is not None:
            for record in self.cache.entries_with_name(question.name):
                if question.answered_by(record) and not record.is_expired(now):
                    listener.update_record(self, now, record)
        self.notify_all()

    def remove_listener(self, listener):
        """Removes a listener."""
        try:
            self.listeners.remove(listener)
            self.notify_all()
        except Exception as e:  # TODO stop catching all Exceptions
            log.exception('Unknown error, possibly benign: %r', e)

    def update_record(self, now, rec):
        """Used to notify listeners of new information that has updated
        a record."""
        for listener in self.listeners:
            listener.update_record(self, now, rec)
        self.notify_all()

    def handle_response(self, msg):
        """Deal with incoming response packets.  All answers
        are held in the cache, and listeners are notified."""
        now = current_time_millis()
        for record in msg.answers:
            expired = record.is_expired(now)
            if record in self.cache.entries():
                if expired:
                    self.cache.remove(record)
                else:
                    entry = self.cache.get(record)
                    if entry is not None:
                        entry.reset_ttl(record)
            else:
                self.cache.add(record)

        for record in msg.answers:
            self.update_record(now, record)

    def handle_query(self, msg, addr, port):
        """Deal with incoming query packets.  Provides a response if
        possible."""
        out = None

        # Support unicast client responses
        #
        if port != _MDNS_PORT:
            out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA, multicast=False)
            for question in msg.questions:
                out.add_question(question)

        for question in msg.questions:
            if question.type == _TYPE_PTR:
                if question.name == "_services._dns-sd._udp.local.":
                    for stype in self.servicetypes.keys():
                        if out is None:
                            out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
                        out.add_answer(msg, DNSPointer(
                            "_services._dns-sd._udp.local.", _TYPE_PTR,
                            _CLASS_IN, _DNS_TTL, stype))
                for service in self.services.values():
                    if question.name == service.type:
                        if out is None:
                            out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
                        out.add_answer(msg, DNSPointer(
                            service.type, _TYPE_PTR,
                            _CLASS_IN, _DNS_TTL, service.name))
            else:
                try:
                    if out is None:
                        out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)

                    # Answer A record queries for any service addresses we know
                    if question.type in (_TYPE_A, _TYPE_ANY):
                        for service in self.services.values():
                            if service.server == question.name.lower():
                                out.add_answer(msg, DNSAddress(
                                    question.name, _TYPE_A,
                                    _CLASS_IN | _CLASS_UNIQUE,
                                    _DNS_TTL, service.address))

                    service = self.services.get(question.name.lower(), None)
                    if not service:
                        continue

                    if question.type in (_TYPE_SRV, _TYPE_ANY):
                        out.add_answer(msg, DNSService(
                            question.name, _TYPE_SRV, _CLASS_IN | _CLASS_UNIQUE,
                            _DNS_TTL, service.priority, service.weight,
                            service.port, service.server))
                    if question.type in (_TYPE_TXT, _TYPE_ANY):
                        out.add_answer(msg, DNSText(
                            question.name, _TYPE_TXT, _CLASS_IN | _CLASS_UNIQUE,
                            _DNS_TTL, service.text))
                    if question.type == _TYPE_SRV:
                        out.add_additional_answer(DNSAddress(
                            service.server, _TYPE_A, _CLASS_IN | _CLASS_UNIQUE,
                            _DNS_TTL, service.address))
                except Exception:  # TODO stop catching all Exceptions
                    self.log_exception_warning()

        if out is not None and out.answers:
            out.id = msg.id
            self.send(out, addr, port)

    def send(self, out, addr=_MDNS_ADDR, port=_MDNS_PORT):
        """Sends an outgoing packet."""
        packet = out.packet()
        if len(packet) > _MAX_MSG_ABSOLUTE:
            self.log_warning_once("Dropping %r over-sized packet (%d bytes) %r",
                                  out, len(packet), packet)
            return
        log.debug('Sending %r (%d bytes) as %r...', out, len(packet), packet)
        for s in self._respond_sockets:
            if self._GLOBAL_DONE:
                return
            try:
                bytes_sent = s.sendto(packet, 0, (addr, port))
            except Exception:   # TODO stop catching all Exceptions
                # on send errors, log the exception and keep going
                self.log_exception_warning()
            else:
                if bytes_sent != len(packet):
                    self.log_warning_once(
                        '!!! sent %d out of %d bytes to %r' % (
                            bytes_sent, len(packet)), s)

    def close(self):
        """Ends the background threads, and prevent this instance from
        servicing further queries."""
        if not self._GLOBAL_DONE:
            self._GLOBAL_DONE = True
            # remove service listeners
            self.remove_all_service_listeners()
            self.unregister_all_services()

            # shutdown recv socket and thread
            self.engine.del_reader(self._listen_socket)
            self._listen_socket.close()
            self.engine.join()

            # shutdown the rest
            self.notify_all()
            self.reaper.join()
            for s in self._respond_sockets:
                s.close()


import atexit
import logging

import paho.mqtt.client as mqtt

logger = logging.getLogger('mqtt')


class MQTT:
    def __init__(self, ip: str, port: int = 1883, user: str = None, cred: str = None):
        import socket
        hostname = socket.gethostname()
        client_id = f'cmdserver-{hostname}'
        logger.info(f'Initialising MQTT client {client_id} to {ip}:{port}')
        self.__client = mqtt.Client(client_id=client_id)
        self.__client.on_connect = self.__on_connect
        self.__client.on_message = self.__on_message
        self.__client.on_disconnect = self.__on_disconnect
        if user and cred:
            self.__client.username_pw_set(user, password=cred)
        self.__client.enable_logger(logger)
        self.__client.connect_async(ip, port, 60)
        self.__client.loop_start()
        import atexit
        atexit.register(self.shutdown)

    def shutdown(self):
        logger.info('Shutting down MQTT client')
        self.__client.loop_stop(force=True)

    def __on_connect(self, client, userdata, flags, rc):
        logger.info(f'Connected to MQTT [result: {rc}]')

    def __on_message(self, client, userdata, msg):
        logger.info(f'Received from {msg.topic} {msg.payload}')

    def __on_disconnect(self, client, userdata, rc):
        logger.warning(f'Disconnected from MQTT [result: {rc}]')

    def publish(self, source: str, payload):
        logger.info(f'Publishing {source} -- {payload}')
        self.__client.publish(f'cmdserver/{source}', qos=1, payload=payload, retain=True)

    def online(self, source: str):
        self.publish(f'{source}/available', 'online')

    def offline(self, source: str):
        self.publish(f'{source}/available', 'offline')



from datetime import datetime
import enum
import logging
import socket

import select
import time

from cmdserver.jvccommands import WriteOnly, BinaryData, ReadOnly, NoVerify

PJ_ACK = b'PJACK'

PJ_REQ = b'PJREQ'

PJ_OK = b'PJ_OK'

DEFAULT_PORT = 20554

logger = logging.getLogger('jvc')


class Error(Exception):
    """Unspecified error"""
    pass


class BadData(Exception):
    def __init__(self, expected, actual):
        self.__expected = expected
        self.__actual = actual

    def __str__(self):
        return f"Expected {self.__expected}, received {self.__actual}"


class Closed(Exception):
    """Connection Closed"""
    pass


class Timeout(Exception):
    """Command Timout"""
    pass


class CommandNack(Exception):
    """JVC command not acknowledged"""
    pass


class Header(enum.Enum):
    """JVC command and response headers"""
    operation = b'!'
    reference = b'?'
    response = b'@'
    ack = b'\x06'


UNIT_ID = b'\x89\x01'


END = b'\x0a'


class Protocol:
    """JVC projector protocol, understands how to send commands and handle the responses"""
    def __init__(self, host):
        self.conn = Connection(host=host)
        self.reconnect = False

    def __enter__(self):
        self.conn.__enter__()
        return self

    def __exit__(self, exception, value, traceback):
        self.conn.__exit__(exception, value, traceback)

    def __cmd(self, cmdtype, cmd, sendrawdata=None, acktimeout=1):
        """Send command and optional raw data and wait for acks"""
        logger.debug(f"  > Cmd:{cmdtype} {cmdtype.value+cmd}")
        assert cmdtype == Header.operation or cmdtype == Header.reference

        retry_count = 1

        def do_send(f):
            nonlocal retry_count
            try:
                f()
            except Closed:
                self.reconnect = True
                if retry_count:
                    logger.warning(f"Connection closed, retry {retry_count}")
                    retry_count -= 1
                else:
                    raise
            except CommandNack:
                self.reconnect = True
                raise
            except:
                self.reconnect = True
                logger.exception(f"Unexpected exception for Cmd:{cmdtype} {cmdtype.value+cmd}")
                raise

        while True:
            if self.reconnect:
                self.conn.reconnect()
                self.reconnect = False
            do_send(lambda: self.__send_cmd(acktimeout, cmd, cmdtype))
            do_send(lambda: self.__send_raw_data(cmd, cmdtype, sendrawdata))
            break

    def __send_raw_data(self, cmd, cmdtype, sendrawdata):
        if sendrawdata is not None:
            self.conn.send(sendrawdata)
            if self.conn.expect(Header.ack.value + UNIT_ID + cmd[:2] + END, timeout=20) == -1:
                raise CommandNack(f"Data not acknowledged [{cmdtype} {len(sendrawdata)}]")

    def __send_cmd(self, acktimeout, cmd, cmdtype):
        data = cmdtype.value + UNIT_ID + cmd + END
        self.conn.send(data)
        if self.conn.expect(Header.ack.value + UNIT_ID + cmd[:2] + END, timeout=acktimeout) == -1:
            raise CommandNack(f"Command not acknowledged [{cmdtype} {data}]")

    def cmd_op(self, cmd, **kwargs):
        """Send operation command"""
        self.__cmd(Header.operation, cmd, **kwargs)

    def cmd_ref(self, cmd, **kwargs):
        """Send reference command and retrieve response"""
        self.__cmd(Header.reference, cmd, **kwargs)
        data = self.conn.recv()
        header = Header.response.value + UNIT_ID + cmd[:2]
        if not data.startswith(header):
            raise Exception('Expected response header', header, data)
        if not data.endswith(END):
            raise Exception('Expected END', END, data)
        res = data[len(header):-1]
        logger.debug(f"  < Response: {res}")
        return res

    def cmd_ref_bin(self, cmd, **kwargs):
        """Send command and retrieve binary response"""
        self.__cmd(Header.reference, cmd, **kwargs)
        try:
            res = self.conn.recv(timeout=10)
        except Timeout:
            self.reconnect = True
            raise
        logger.debug(f"  < Response:{res}")
        return res


class Connection:
    """JVC projector network connection, handles low level socket comms and connection initialisation """
    def __init__(self, host, port=DEFAULT_PORT):
        self.__socket = None
        self.__port = port
        self.__host = host
        self.__close_time = 0.0

    def connect(self):
        """Open network connection to projector and perform handshake"""
        if self.__socket is None:
            self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            request_time = time.time()
            to_wait = 1.0 - (request_time - self.__close_time)
            if 1 > to_wait > 0:
                close_time_fmt = datetime.fromtimestamp(self.__close_time).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
                logger.info(f"Waiting {to_wait:.3f}s to reopen socket [closed at: {close_time_fmt}]")
                time.sleep(to_wait)
            try:
                logger.info(f"Connecting to {self.__host}:{self.__port}")
                self.__socket.settimeout(1)
                self.__socket.connect((self.__host, self.__port))
                logger.info(f"Connected to {self.__host}:{self.__port}")
                self.__init_pj()
            except socket.timeout:
                raise Timeout(f"Connection failed on timeout [{self.__host}:{self.__port}]")
            except Exception as err:
                raise Error(f"Connection failed [{self.__host}:{self.__port}]", err)

    def __init_pj(self):
        logger.debug(f">> Protocol init")
        self.expect(PJ_OK)
        self.send(PJ_REQ)
        self.expect(PJ_ACK)
        logger.debug(f"<< Protocol init")

    def __enter__(self):
        while True:
            try:
                self.connect()
            except Exception as e:
                logger.exception('Unexpected failure')
                self.close()
                raise e
            break

        return self

    def close(self, fail=True):
        if self.__socket is not None:
            logger.info(f"Closing connection to {self.__host}:{self.__port}")
            try:
                self.__socket.close()
                logger.info(f"Closed connection to {self.__host}:{self.__port}")
            except Exception as e:
                if fail is True:
                    raise e
                else:
                    logger.exception(f"Unable to close connection to {self.__host}:{self.__port}")
            finally:
                self.__socket = None
                self.__close_time = time.time()

    def __exit__(self, exception, value, traceback):
        self.close()

    def reconnect(self):
        self.close()
        self.connect()

    def send(self, data):
        if self.__socket:
            logger.debug(f"Sending {data} to {self.__host}:{self.__port}")
            try:
                self.__socket.send(data)
            except ConnectionAbortedError as err:
                raise Closed(err)
        else:
            raise Error('Unable to send, not connected')

    def recv(self, limit=1024, timeout=1):
        if timeout:
            ready = select.select([self.__socket], [], [], timeout)
            if not ready[0]:
                raise Timeout(f"{timeout} second timeout expired")
        data = self.__socket.recv(limit)
        if not len(data):
            raise Closed('Connection closed by projector')
        logger.debug(f"< Received:{data}")
        return data

    def expect(self, expected, timeout=1):
        """Receive data and compare it against expected data"""
        try:
            actual = self.recv(len(expected), timeout)
        except Timeout:
            return -1
        if actual != expected:
            raise BadData(expected, actual)
        else:
            return 0


class CommandExecutor:
    """ Provides ability to execute specific commands """
    def __init__(self, host):
        self.conn = Protocol(host)

    def __enter__(self):
        self.conn.__enter__()
        return self

    def __exit__(self, exception, value, traceback):
        self.conn.__exit__(exception, value, traceback)

    def connect(self):
        self.__enter__()

    def disconnect(self, fail=True):
        self.conn.conn.close(fail=fail)

    def get(self, cmd):
        """Send reference command and convert response"""
        if isinstance(cmd.value, bytes):
            raise NotImplementedError('Get is not implemented for {}'.format(cmd.name))
        cmdcode, valtype = cmd.value
        if issubclass(valtype, WriteOnly):
            raise TypeError('{} is a write only command'.format(cmd.name))
        response = self.conn.cmd_ref_bin(cmdcode) if issubclass(valtype, BinaryData) else self.conn.cmd_ref(cmdcode)
        return valtype(response)

    def set(self, cmd, val, verify=True):
        """Send operation command"""
        cmdcode, valtype = cmd.value
        assert not issubclass(valtype, ReadOnly), '{} is a read only command'.format(cmd)
        val = valtype(val)
        assert(isinstance(val, valtype)), '{} is not {}'.format(val, valtype)
        if issubclass(valtype, BinaryData):
            self.conn.cmd_op(cmdcode, sendrawdata=val.value)
        else:
            self.conn.cmd_op(cmdcode+val.value, acktimeout=5)

        if not verify or issubclass(valtype, NoVerify):
            return

        verify_val = self.get(cmd)
        if verify_val != val:
            raise CommandNack('Verify error: ' + cmd.name, val, verify_val)



import logging
import re
import socket
import threading
import time
from typing import Optional

from cmdserver.zeroconf import Zeroconf, ServiceBrowser
from cmdserver.config import Config
from cmdserver.mqtt import MQTT

logger = logging.getLogger('tivocontroller')

# Mapping of commands to remote codes
CODES = {
    # teleport
    'Home': 'TIVO',
    'My Shows': 'NOWSHOWING',
    'LiveTV': 'LIVETV',
    'Info': 'INFO',
    'Guide': 'GUIDE',
    'Back': 'BACK',
    'Thumbs Up': 'THUMBSUP',
    'Thumbs Down': 'THUMBSDOWN',
    # navigation
    'Up': 'UP',
    'Down': 'DOWN',
    'Left': 'LEFT',
    'Right': 'RIGHT',
    'Return': 'SELECT',
    'Ch+': 'CHANNELUP',
    'Ch-': 'CHANNELDOWN',
    # Play, Pause, FF, Rew
    'Stop': 'STOP',
    'Play': 'PLAY',
    'Rewind': 'REVERSE',
    'Pause': 'PAUSE',
    'FF': 'FORWARD',
    # Skip Back, Slow Motion, Skip Forward
    'Replay': 'REPLAY',
    'Slow': 'SLOW',
    'Advance': 'ADVANCE',
    # Record
    'Record': 'RECORD',
    # colour keys
    'Red': 'ACTION_A',
    'Green': 'ACTION_B',
    'Yellow': 'ACTION_C',
    'Blue': 'ACTION_D',
    # channel selection
    'Clear': 'CLEAR',
    'Enter': 'ENTER',
    # subtitles
    'Subtitles Off': 'CC_OFF',
    'Subtitles On': 'CC_ON'
}

# Named symbols for direct text input -- these work with IRCODE and
# KEYBOARD commands
SYMBOLS = {'-': 'MINUS', '=': 'EQUALS', '[': 'LBRACKET',
           ']': 'RBRACKET', '\\': 'BACKSLASH', ';': 'SEMICOLON',
           "'": 'QUOTE', ',': 'COMMA', '.': 'PERIOD', '/': 'SLASH',
           '`': 'BACKQUOTE', ' ': 'SPACE', '1': 'NUM1', '2': 'NUM2',
           '3': 'NUM3', '4': 'NUM4', '5': 'NUM5', '6': 'NUM6',
           '7': 'NUM7', '8': 'NUM8', '9': 'NUM9', '0': 'NUM0'}

# When in shift mode (with KEYBOARD command only), the same names
# map to a different set of symbols
SHIFT_SYMS = {'_': 'MINUS', '+': 'EQUALS', '{': 'LBRACKET',
              '}': 'RBRACKET', '|': 'BACKSLASH', ':': 'SEMICOLON',
              '"': 'QUOTE', '<': 'COMMA', '>': 'PERIOD', '?': 'SLASH',
              '~': 'BACKQUOTE', '!': 'NUM1', '@': 'NUM2', '#': 'NUM3',
              '$': 'NUM4', '%': 'NUM5', '^': 'NUM6', '&': 'NUM7',
              '*': 'NUM8', '(': 'NUM9', ')': 'NUM0'}

# generated from https://www.virginmedia.com/content/dam/virginmedia/dotcom/documents/Redwood/ChannelGuide_June2019.pdf
# linked from https://www.virginmedia.com/virgin-tv-edit/tips-and-tricks/virgin-tv-channel-guide
CHANNELS = {
    '100': 'Virgin TV Showcase',
    '101': 'BBC One HD',
    '102': 'BBC Two HD',
    '103': 'ITV1 HD & STV HD',
    '104': 'Channel 4/HD & S4C HD',
    '105': 'Channel 5 HD',
    '106': 'E4 HD',
    '107': 'BBC Three HD & BBC Four HD',
    '108': 'BBC Four HD/Channel 4 HD Wales',
    '109': 'Sky Showcase HD',
    '110': 'Sky Witness HD',
    '111': 'Sky Max HD',
    '112': 'Sky Comedy HD',
    '113': 'BBC Three HD Scotland & Wales',
    '114': 'alibi HD',
    '115': 'ITV2 HD',
    '116': 'Drama HD',
    '117': 'ITV3 HD',
    '118': 'ITV4 HD',
    '119': 'ITVBe HD',
    '120': '5 USA',
    '121': 'Sky Crime HD',
    '122': 'Sky Sci-Fi HD',
    '123': 'Sky Arts HD',
    '124': 'GOLD HD',
    '125': 'W HD',
    '126': '5STAR',
    '127': 'Dave HD',
    '128': 'Really',
    '129': 'Yesterday HD',
    '130': '5ACTION HD',
    '131': 'Sky HISTORYHD',
    '132': 'Comedy Central HD',
    '133': 'Crime + Investigation HD',
    '134': 'MTV HD',
    '135': 'Sky Mix',
    '136': 'Together',
    '137': 'Quest HD',
    '138': '5Select',
    '139': 'Challenge',
    '140': 'Sky Arts',
    '143': '4Seven HD',
    '144': 'E4 Extra',
    '147': 'More4 HD',
    '148': 'CBS Reality',
    '149': 'Legend',
    '150': 'That’s TV',
    '154': 'Comedy Central Extra',
    '156': 'Sky Replay',
    '159': 'Regional Channels',
    '160': 'E! HD',
    '161': 'BBC Alba',
    '162': '5USA',
    '164': 'S4C HD',
    '165': 'TLC HD',
    '166': 'Investigation Discovery',
    '168': 'Quest Red',
    '169': 'DMAX',
    '170': 'GREAT TV',
    '171': 'Horror Xtra',
    '172': 'RealityXtra',
    '174': 'BLAZE',
    '175': 'Virgin TV Ultra HD',
    '176': 'Eden HD',
    '177': 'Discovery HD',
    '178': 'Animal Planet HD',
    '179': 'Discovery Science',
    '180': 'Discovery Turbo',
    '181': 'Discovery History',
    '182': 'National Geographic WILD HD',
    '183': 'National Geographic HD',
    '186': 'Sky HISTORY2HD',
    '187': 'PBS America',
    '188': 'Sky Documentaries HD',
    '189': 'Sky Nature HD',
    '190': 'Food Network',
    '191': 'HGTV',
    '192': 'God TV',
    '360': 'Welcome',
    '204': 'Netflix',
    '205': 'Prime Video',
    '220': 'Inside Crime',
    '221': 'Real Wild',
    '222': 'Mystery TV',
    '223': 'Haunt TV',
    '224': 'History Hit',
    '230': 'Homes Under The Hammer',
    '231': 'Great British Menu',
    '232': 'Tastemade',
    '242': 'NextUp Live Comedy',
    '243': 'The Chat Show Channel',
    '244': 'Baywatch',
    '250': 'Deal Or No Deal',
    '251': 'Fear Factor',
    '252': 'Wipeout Xtra',
    '991': 'BBC Red Button HD',
    '998': 'Virgin TV Highlights',
    '999': 'Virgin TV Ultra HD',
    '280': 'MTV Music',
    '281': 'MTV Live HD',
    '282': 'MTV Hits',
    '283': 'MTV 90s',
    '284': 'MTV 80s',
    '285': 'The Box',
    '286': '4Music',
    '288': 'Kiss',
    '289': 'Magic',
    '290': 'Kerrang!',
    '291': 'Vevo',
    '292': 'Clubland TV',
    '293': 'Now 70s',
    '294': 'Now 80s',
    '295': 'Now Rock',
    '296': 'That’s 60s',
    '303': 'ITV1 +1 & STV +1',
    '304': 'Channel 4 +1',
    '305': 'Channel 5 +1',
    '306': 'E4 +1',
    '310': 'Sky Witness +1',
    '314': 'alibi +1',
    '315': 'ITV2 +1',
    '316': 'Drama HD +1',
    '317': 'ITV3 +1',
    '318': 'ITV4 +1',
    '319': 'ITVBe +1',
    '320': '5 USA +1',
    '321': 'Sky Crime +1',
    '324': 'GOLD +1',
    '325': 'W +1',
    '326': '5STAR +1',
    '327': 'Dave ja vu',
    '329': 'Yesterday HD +1',
    '331': 'Sky HISTORY +1',
    '332': 'Comedy Central +1',
    '333': 'Crime + Investigation +1',
    '337': 'Quest +1',
    '347': 'More4 +1',
    '348': 'CBS Reality +1',
    '365': 'TLC +1',
    '366': 'Investigation Discovery +1',
    '369': 'DMAX +1',
    '371': 'HorrorXtra +1',
    '376': 'Eden +1',
    '377': 'Discovery +1',
    '378': 'Animal Planet +1',
    '379': 'Discovery Science +1',
    '381': 'Discovery History +1',
    '383': 'National Geographic +1',
    '391': 'HGTV +1',
    '400': 'Virgin Movies & Store',
    '401': 'Sky Cinema Premiere HD',
    '402': 'Sky Cinema Select HD',
    '403': 'Sky Cinema Hits HD',
    '404': 'Sky Cinema Greats HD',
    '406': 'Sky Cinema Family HD',
    '407': 'Sky Cinema Action HD',
    '408': 'Sky Cinema Comedy HD',
    '409': 'Sky Cinema Thriller HD',
    '410': 'Sky Cinema Drama HD',
    '411': 'Sky Cinema Sci-Fi & Horror HD',
    '412': 'Sky Cinema Animation HD',
    '419': 'Movies 24',
    '420': 'Movies 24 +1',
    '424': 'Great! Romance',
    '425': 'Great! Movies',
    '426': 'Great! Action',
    '428': 'Film4 HD',
    '430': 'Film4 +1',
    '445': 'Talking Pictures TV',
    '501': 'Sky Sports Main Event/HD',
    '502': 'Sky Sports Premier League/HD',
    '503': 'Sky Sports Football/HD',
    '504': 'Sky Sports Cricket/HD',
    '505': 'Sky Sports Golf/HD',
    '506': 'Sky Sports F1®/HD',
    '507': 'Sky Sports Action/HD',
    '508': 'Sky Sports Arena/HD',
    '509': 'Sky Sports News HD',
    '510': 'Sky Sports Mix HD',
    '519': 'Sky Sports Racing HD',
    '521': 'Eurosport 1 HD',
    '522': 'Eurosport 2 HD',
    '526': 'MUTV',
    '527': 'TNT Sports 1 HD',
    '528': 'TNT Sports 2 HD',
    '529': 'TNT Sports 3 HD',
    '530': 'TNT Sports 4 HD',
    '531': 'TNT Sports Ultimate',
    '532': 'Sky Sports Main Event UHD',
    '533': 'Sky Sports F1® UHD',
    '534': 'Sky Sports UHD',
    '535': 'Sky Sports UHD 2',
    '536': 'Racing TV HD',
    '544': 'LFC TV HD',
    '551': 'Viaplay Sports 1 HD',
    '552': 'Viaplay Sports 2 HD',
    '553': 'Viaplay Xtra',
    '601': 'BBC News HD',
    '602': 'Sky News/HD',
    '604': 'GB News',
    '605': 'BBC Parliament',
    '606': 'TalkTV',
    '609': 'Bloomberg HD',
    '613': 'CNBC',
    '614': 'NBC News NOW HD',
    '620': 'Euronews',
    '621': 'NDTV 24x7',
    '622': 'Al Jazeera English',
    '624': 'France 24 English HD',
    '625': 'NHKworld HD',
    '700': 'Kids TV On Demand',
    '701': 'CBBC HD',
    '702': 'CBeebies HD',
    '703': 'Baby TV',
    '704': 'Cartoon Network HD',
    '702': 'CBeebies HD',
    '703': 'Baby TV',
    '704': 'Cartoon Network HD',
    '705': 'Cartoon Network +1',
    '706': 'Cartoonito',
    '707': 'Sky Kids HD',
    '712': 'Nickelodeon HD',
    '713': 'Nick +1',
    '715': 'Nick Jr.',
    '716': 'Nick Jr. Too',
    '717': 'Nicktoons',
    '730': 'Boomerang',
    '731': 'Boomerang +1',
    '736': 'Pop',
    '737': 'Tiny Pop',
    '740': 'QVC HD',
    '741': 'QVC Beauty',
    '742': 'QVC Style HD',
    '743': 'QVC Extra',
    '748': 'Create and Craft',
    '755': 'Gems TV',
    '757': 'TJC HD',
    '801': 'Utsav Gold HD',
    '802': 'Utsav Bharat',
    '803': 'Utsav Plus HD',
    '805': 'SONY TV HD',
    '806': 'SONY MAX HD',
    '807': 'SONY SAB',
    '808': 'SONY MAX 2',
    '809': 'Zee TV HD',
    '810': 'Zee Cinema HD',
    '815': 'B4U Movies',
    '816': 'B4U Music',
    '825': 'Colors Gujarati',
    '826': 'Colors TV HD',
    '827': 'Colors Rishtey',
    '828': 'Colors Cineplex',
    '829': 'NDTV Good Times',
    '831': 'Al Jazeera Arabic',
    '838': 'Islam Channel',
    '839': 'Islam Channel Urdu',
    '159': 'Regional Channels',
    '861': 'BBC One London HD',
    '862': 'BBC One Scotland HD',
    '863': 'BBC One NI HD',
    '864': 'BBC One Wales HD',
    '865': 'BBC Two England HD',
    '870': 'ITV1'
}


class Tivo:

    def __init__(self, tivo: dict, mqtt: MQTT):
        self.__sock = None
        self.__tivo = tivo
        self.__mqtt = mqtt
        self.__messages = [''] * 5
        self.__message_idx = 0
        self.current_channel = ''
        self.current_channel_num = -1
        if ':' in tivo['address']:
            address, port = tivo['address'].split(':')
            try:
                port = int(port)
                self.__tivo['address'] = address
                self.__tivo['port'] = port
            except:
                pass
        self.__connect()

    @property
    def name(self):
        return self.__tivo['name']

    @property
    def port(self):
        return self.__tivo['port']

    @property
    def address(self):
        return self.__tivo['address']

    @property
    def version(self):
        return self.__tivo['version']

    def get_tivo_properties(self):
        return {**self.__tivo}

    @property
    def connected(self):
        return self.__sock is not None

    def __connect(self):
        """ Connect to the TiVo within five seconds or report error. """
        self.__messages = [''] * 5
        self.__message_idx = 0
        try:
            self.__sock = socket.socket()
            self.__sock.settimeout(5)
            self.__sock.connect((self.address, self.port))
            self.__sock.settimeout(None)
            if self.__mqtt:
                self.__mqtt.online(f'tivo/{self.name}')
            self.__status_thread = threading.Thread(name='TivoStatusReader', target=self.__read_from_socket,
                                                    daemon=True)
            self.__status_thread.start()
        except Exception as e:
            self.__record_message('Unable to connect - ' + str(e))
            self.disconnect()

    def disconnect(self):
        if self.__sock:
            self.__record_message('Stopping on disconnect')
            self.__sock.close()
            self.__record_message('Disconnected')
            self.__sock = None
        else:
            logger.info(f"Ignoring disconnect {self.name} has no socket")
        if self.__mqtt:
            self.__mqtt.offline(f'tivo/{self.name}')

    def __send(self, message):
        """ The core output function. Re-connect if necessary, send message, sleep, and check for errors. """
        if not self.__sock:
            self.__connect()
        try:
            self.__sock.sendall(message.encode())
            time.sleep(0.1)
        except Exception as e:
            logger.error(e)
            self.__record_message('Failed to send ' + str(message) + ' - ' + str(e))
            self.disconnect()
            raise e

    def send_ir(self, sent, *codes):
        """ Expand a command sequence for send(). """
        for each in codes:
            if type(each) == list:
                self.send_ir(sent, *each)
            else:
                if each in CODES:
                    actual = CODES[each]
                    if type(actual) == list:
                        self.send_ir(sent, *actual)
                    else:
                        to_send = 'IRCODE %s\r' % actual
                        sent.append(to_send)
                        self.__send(to_send)
                else:
                    raise ValueError('Unknown IR Code ' + each)
        return {'channel': self.current_channel, 'sent': sent}

    def set_channel(self, channel):
        """ Sends a single SETCH. """
        toSend = 'SETCH %s\r' % channel
        self.__send(toSend)
        return {'channel': self.current_channel, 'sent': [toSend]}

    def send_text(self, text):
        """ Expand a KEYBOARD command sequence for send(). """
        sent = []
        for ch in text:
            if 'A' <= ch <= 'Z':
                self.__send_keyboard(sent, 'LSHIFT', ch)
            elif 'a' <= ch <= 'z':
                self.__send_keyboard(sent, ch.upper())
            elif ch in SYMBOLS:
                self.__send_keyboard(sent, SYMBOLS[ch])
            elif ch in SHIFT_SYMS:
                self.__send_keyboard(sent, 'LSHIFT', SHIFT_SYMS[ch])
        return {'channel': self.current_channel, 'sent': sent}

    def __send_keyboard(self, sent, *codes):
        for each in codes:
            to_send = 'KEYBOARD %s\r' % each
            sent.append(to_send)
            self.__send(to_send)

    def __record_message(self, msg):
        idx = self.__message_idx % 5
        self.__message_idx += 1
        self.__messages[idx] = msg

    def __read_from_socket(self):
        """ Read incoming messages from the socket in a separate thread.
        """
        logger.info(f"[{self.name}] Initialising status reader")
        while self.__sock is not None:
            logger.info(f'[{self.name}] Reading from socket')
            try:
                status = self.__sock.recv(80).decode()
            except Exception as e:
                status = str(e)
                logger.warning(f"Failed to read - {status}")
            status = status.strip().title()
            if not status:
                self.__record_message('No status read from socket')
                self.disconnect()
            else:
                last_ch = self.current_channel
                self.current_channel = status
                if status.startswith('Ch_Status'):
                    try:
                        self.current_channel_num = int(status.split()[1])
                    except:
                        self.current_channel_num = -1
                else:
                    self.current_channel_num = -1
                if last_ch != self.current_channel and self.__mqtt:
                    ch_name = CHANNELS.get(str(self.current_channel_num), 'Unknown')
                    self.__mqtt.publish(f'tivo/{self.name}/state', ch_name)
        logger.info(f"[{self.name}] Exiting status reader")

    @property
    def messages(self):
        return self.__messages


class TivoController(object):

    def __init__(self, cfg: Config, mqtt: Optional[MQTT]):
        if cfg.tivo:
            logger.info(f"Using static tivo config: {cfg.tivo}")
            self.__tivos = [Tivo(cfg.tivo, mqtt)]
        elif cfg.find_tivo:
            logger.info(f"Discovering Tivos")
            self.__tivos = [Tivo(t, mqtt) for t in self.__find_tivos()]
            self.__ping()
        else:
            logger.info(f"Tivo support disabled")
            self.__tivos = []

    def __ping(self, delay=15):
        time.sleep(delay)
        worker = threading.Thread(name='TivoPinger', target=self.__ping0, daemon=True)
        worker.start()

    def __ping0(self):
        try:
            tivos = self.__find_tivos()
            logger.debug(f"Found {len(tivos)} tivos")
            for new_tivo in tivos:
                existing_tivo = next((t for t in self.__tivos if t.name == new_tivo['name']), None)
                if existing_tivo is None:
                    logger.info(f"Connecting new Tivo {new_tivo['name']} @ {new_tivo['address']}:{new_tivo['port']}")
                    self.__tivos.append(Tivo(new_tivo))
            for old_tivo in self.__tivos:
                dead_tivo = next((t for t in tivos if t['name'] == old_tivo.name), None)
                if dead_tivo is None:
                    logger.info(f"Disconnecting old Tivo {old_tivo.name} @ {old_tivo.address}:{old_tivo.port}")
                    old_tivo.disconnect()
        except:
            logger.exception('Unexpected failure during ping')
        finally:
            self.__ping()

    def has_tivo(self, tivo_name: str) -> bool:
        return True if next((t for t in self.__tivos if t.name == tivo_name), None) else False

    def get_tivo(self, tivo_name: str) -> Optional[Tivo]:
        return next((t for t in self.__tivos if t.name == tivo_name), None)

    def send(self, tivo_name, cmd_type, command):
        tivo = self.get_tivo(tivo_name)
        if tivo is not None:
            if cmd_type == 'keyboard':
                return tivo.send_text(command)
            elif cmd_type == 'ir':
                return tivo.send_ir([], command)
            elif cmd_type == 'setch':
                return tivo.set_channel(command)
            elif cmd_type == 'setchno':
                ch_num = next((k for k, v in CHANNELS.items() if v == command), '')
                if ch_num:
                    return tivo.set_channel(ch_num)
                else:
                    return ValueError(f'Unknown channel {command}')
            else:
                raise ValueError('Unknown command type ' + cmd_type)
        else:
            raise ValueError('Unknown tivo ' + tivo_name)

    @property
    def tivos(self):
        return [
            {
                **t.get_tivo_properties(),
                **{'connected': t.connected, 'messages': t.messages, 'channel': t.current_channel,
                   'channelNumber': t.current_channel_num}
            }
            for t in self.__tivos
        ]

    def __find_tivos(self):
        class ZCListener:
            def __init__(self, names):
                self.names = names

            def remove_service(self, server, type, name):
                self.names.remove(name)

            def add_service(self, server, type, name):
                self.names.append(name)

        REMOTE = '_tivo-remote._tcp.local.'

        tivos = []
        tivos_names = []

        # Get the names of TiVos offering network remote control
        serv = Zeroconf()
        try:
            browser = ServiceBrowser(serv, REMOTE, ZCListener(tivos_names))
            # Give them a second to respond
            time.sleep(1)

            # For proxied TiVos, remove the original
            # for t in tivos_names[:]:
            #     if t.startswith('Proxy('):
            #         try:
            #             t = t.replace('.' + REMOTE, '')[6:-1] + '.' + REMOTE
            #             tivos_names.remove(t)
            #         except:
            #             pass

            # Now get the addresses -- this is the slow part
            swversion = re.compile(r'(\d*.\d*)').findall
            for t in tivos_names:
                s = serv.get_service_info(REMOTE, t)
                if s:
                    name = t.replace('.' + REMOTE, '')
                    address = socket.inet_ntoa(s.address)
                    try:
                        versionProperty = s.properties.get(str.encode('swversion'))
                        if (versionProperty):
                            version = float(swversion(versionProperty.decode())[0])
                        else:
                            version = 0.0
                    except Exception as e:
                        print(e)
                        version = 0.0
                    tivos.append({
                        'name': name,
                        'port': s.port,
                        'version': version,
                        'address': address
                    })

            # For proxies with numeric names, remove the original
            # for name, tivo in tivos.items():
            #     if name.startswith('Proxy('):
            #         address = name.replace('.' + REMOTE, '')[6:-1]
            #         if address in tivos_rev:
            #             tivos.pop(tivos_rev[address])
            return tivos
        except Exception as e:
            print(e)
            return tivos
        finally:
            serv.close()



import functools
import json
import logging
import re
from typing import Tuple, List, Optional

import pymcws

from plumbum import local
from pymcws import MediaServer, Zone
from requests import ConnectTimeout
from twisted.internet import task, threads
from twisted.internet.defer import Deferred
from urllib3.exceptions import MaxRetryError

from cmdserver.config import Config
from cmdserver.ws import WsServer

DEFAULT_ACTIVE_COMMAND = 'Music'

logger = logging.getLogger('infoprovider')


class InfoProvider:

    VOLUME_PATTERN = re.compile(r'.*\(([-+][0-9]+\.[0-9]) dB\)')

    def __init__(self, config: Config, ws_server: WsServer):
        self.__ws_server = ws_server
        self.__interval = float(config.mcws.get('interval', 0.25))
        self.__launcher = local[config.playingNowExe] if config.playingNowExe else None
        self.__default_playing_now_id = None
        for key, value in config.commands.items():
            if 'playingNowId' in value:
                is_default = value.get('defaultPlayingNowId', False)
                if is_default is True:
                    self.__default_playing_now_id = value['playingNowId']
        self.__by_playing_now_id = {value['playingNowId']: value['title']
                                    for key, value in config.commands.items() if 'playingNowId' in value}
        self.__ms: MediaServer = MediaServer('localhost', config.mcws.get('user', None), config.mcws.get('pass', None))
        # make sure all calls in pymcws have a timeout not just connection checks
        self.__ms.session.request = functools.partial(self.__ms.session.request, timeout=2.5)
        self.__ms_mac: str = config.mcws.get('mac', '')
        self.__ms.port = int(config.mcws.get('port', 52199))
        self.__ms.local_ip_list = config.mcws.get('ip', '127.0.0.1')
        self.__ms.local_ip = self.__ms.local_ip_list
        self.__token = None
        self.__current_state = {}
        self.__refresh_task = task.LoopingCall(self.refresh)
        self.__deferred = self.__refresh_task.start(self.__interval)

    @property
    def info(self):
        return self.__current_state

    def __get_token(self) -> str:
        if self.__token is None:
            response = self.__ms.send_request("Authenticate")
            response.raise_for_status()
            from pymcws.utils import transform_unstructured_response
            content = transform_unstructured_response(response)
            self.__token = content['Token']
        return self.__token

    def refresh(self) -> Deferred:
        return threads.deferToThread(self.__async_refresh)

    def __async_refresh(self):
        try:
            zones, active_zone = get_zones(self.__ms)
            playback_info = get_playback_info(self.__ms, active_zone)
            zones_data = {}
            for z in zones:
                if z == active_zone:
                    zd = self.__zone_to_dict(z, playback_info)
                else:
                    zd = self.__zone_to_dict(z, None)
                zones_data[z.id] = zd
            self.__current_state = {
                'config': {
                    'host': self.__ms.local_ip,
                    'port': self.__ms.port,
                    'token': self.__get_token(),
                    'ssl': False,
                    'alive': True
                },
                'zones': zones_data
            }
            active_command = self.get_active_command(self.__current_state['zones'].get(active_zone.id, None))
            self.__current_state['playingCommand'] = {'active': active_command}
            self.__ws_server.broadcast(json.dumps(self.__current_state, ensure_ascii=False))
        except ConnectTimeout as e:
            logger.warning(f"Unable to connect, MC probably sleeping {e.request.method} {e.request.url}")
            self.__broadcast_down()
        except OSError as e:
            if e.errno == 113:
                logger.warning(f"Unable to connect, MC probably sleeping {e.request.method} {e.request.url}")
            else:
                e = e.__context__
                while e is not None and not isinstance(e, OSError):
                    e = e.__context__
                if e is not None and e.errno == 113:
                    logger.warning(f"Unable to connect, MC probably sleeping {e.request.method} {e.request.url}")
                else:
                    logger.exception(f"Unknown OSError when refreshing current state of {self.__ms.address()}")
            self.__broadcast_down()
        except ConnectionRefusedError as e:
            logger.error(f"Unable to connect, MC appears down")
            self.__broadcast_down()
        except:
            logger.exception(f"Unexpected failure to refresh current state of {self.__ms.address()}")
            self.__broadcast_down()

    def __broadcast_down(self):
        was_alive = not self.__current_state or self.__current_state.get('config', {}).get('alive', True) is True
        if was_alive:
            logger.warning(f"Broadcasting MC is down")
        self.__current_state = {
            'config': {
                'host': self.__ms.local_ip,
                'port': self.__ms.port,
                'token': None,
                'ssl': False,
                'alive': False
            },
            'zones': self.__current_state['zones'] if self.__current_state else {},
            'playingCommand': {'active': DEFAULT_ACTIVE_COMMAND}
        }
        self.__ws_server.broadcast(json.dumps(self.__current_state, ensure_ascii=False))

    def wake(self) -> bool:
        if self.__ms_mac:
            from wakeonlan import send_magic_packet
            logger.info(f"Sending magic packet to {self.__ms_mac}")
            send_magic_packet(self.__ms_mac)
            logger.info(f"Sent magic packet to {self.__ms_mac}")
            return True
        else:
            return False

    @staticmethod
    def __zone_to_dict(zone: Zone, playback_info):
        zd = {'name': zone.name, 'id': zone.id, 'active': playback_info is not None}
        pn = None
        if playback_info:
            InfoProvider.__extract_volume(playback_info, zd)
            pn = {
                'artist': playback_info.get('Artist', ''),
                'album': playback_info.get('Album', ''),
                'status': InfoProvider.__extract_status(playback_info),
                'fileKey': playback_info.get('FileKey', ''),
                'positionMillis': float(playback_info.get('PositionMS', 0)),
                'durationMillis': float(playback_info.get('DurationMS', 0)),
                'positionDisplay': playback_info.get('PlayingNowPositionDisplay', ''),
                'imageURL': playback_info.get('ImageURL', ''),
                'name': playback_info.get('Name', ''),
                'externalSource': playback_info.get('Name', '') == 'Ipc',
                'hdr': playback_info.get('HDR Format', ''),
                'ar': playback_info.get('Aspect Ratio', ''),
                'hdr_override': playback_info.get('hdr_override', ''),
            }
            zd['volumeRatio'] = float(playback_info.get('Volume', 0))
        zd['playingNow'] = pn
        return zd

    @staticmethod
    def __extract_volume(playback_info: dict, zd: dict):
        volume_display = playback_info.get('VolumeDisplay', None)
        if not volume_display or volume_display == 'Muted':
            zd['muted'] = True
            zd['volumedb'] = -100.0
        else:
            zd['muted'] = False
            if volume_display == '100%':
                zd['volumedb'] = 0.0
            else:
                m = InfoProvider.VOLUME_PATTERN.match(volume_display)
                if m:
                    zd['volumedb'] = float(m[1])
                else:
                    zd['volumedb'] = -100.0

    def get_active_command(self, zone: Optional[dict]):
        if zone:
            pn = zone.get('playingNow', None)
            if pn and pn['externalSource'] is True:
                if self.__launcher:
                    playing_now = self.__launcher.run(retcode=None)
                    playing_now_id = playing_now[0]
                    logger.debug("playingNow:" + str(playing_now_id))
                    if playing_now_id in self.__by_playing_now_id:
                        return self.__by_playing_now_id[playing_now_id]
                elif self.__default_playing_now_id:
                    return self.__by_playing_now_id[self.__default_playing_now_id]
        return zone['name'] if zone else DEFAULT_ACTIVE_COMMAND

    @staticmethod
    def __extract_status(playback_info):
        state = int(playback_info.get('State', -1))
        if state == 0:
            return 'Stopped'
        elif state == 1:
            return 'Paused'
        elif state == 2:
            return 'Playing'
        else:
            return 'Unknown'


def get_zones(media_server: MediaServer, see_hidden: bool = False) -> Tuple[List[Zone], Zone]:
    see_hidden = "1" if see_hidden else "0"
    payload = {"Hidden": see_hidden}
    response = media_server.send_request("Playback/Zones", payload)
    response.raise_for_status()
    from pymcws.utils import transform_unstructured_response
    content = transform_unstructured_response(response)
    num_zones = int(content["NumberZones"])
    active_zone_id = content['CurrentZoneID']
    active_zone = None
    zones = []
    for i in range(num_zones):
        zone = Zone()
        zone.index = i
        zone.id = content["ZoneID" + str(i)]
        zone.name = content["ZoneName" + str(i)]
        zone.guid = content["ZoneGUID" + str(i)]
        zone.is_dlna = True if (content["ZoneDLNA" + str(i)] == "1") else False
        zones.append(zone)
        if active_zone_id == zone.id:
            active_zone = zone
    if not active_zone:
        active_zone = zones[0]
    return zones, active_zone


def get_playback_info(media_server: MediaServer, active_zone: Zone) -> dict:
    payload = {
        'Fields': 'Aspect Ratio;HDR Format;Playback Info',
        'Zone': active_zone.best_identifier(),
        'ZoneType': active_zone.best_identifier_type()
    }
    response = media_server.send_request("Playback/Info", payload)
    response.raise_for_status()
    from pymcws.utils import transform_unstructured_response
    structured_resp = transform_unstructured_response(response)
    pi = structured_resp.get('Playback Info', '')
    hdr_override = ''
    if pi:
        try:
            pos = pi.index('JRVRProfiles')
            profiles = pi[pos + len('JRVRProfiles') + 1:]
            pos = profiles.index(':')
            profiles_len = profiles[1:pos]
            tokens = [x for x in re.split(r'[()]', profiles[pos + 1: pos + int(profiles_len) + 1]) if x]
            pos = tokens.index('1:3')
            selected = tokens[pos + 1].split(':', 2)
            hdr_override = selected[1]
        except:
            # can't parse so ignoring
            pass
    structured_resp['hdr_override'] = hdr_override
    return structured_resp



import faulthandler
import os
from os import path
from typing import Tuple

from autobahn.twisted.resource import WebSocketResource
from flask import Flask
from flask_restx import Api

from cmdserver.pjcontroller import PJController
from cmdserver.apis import command, commands, playingnow, tivo, tivos, pj, info, version, wake
from cmdserver.commandcontroller import CommandController
from cmdserver.config import Config
from cmdserver.infoprovider import InfoProvider
from cmdserver.tivocontroller import TivoController
from cmdserver.ws import WsServer
from cmdserver.mqtt import MQTT

API_PREFIX = '/api/1'

faulthandler.enable()
if hasattr(faulthandler, 'register'):
    import signal

    faulthandler.register(signal.SIGUSR2, all_threads=True)


def create_app(cfg: Config) -> Tuple[Flask, 'WsServer']:
    ws_server = WsServer()
    info_provider = InfoProvider(cfg, ws_server)
    mqtt = None
    if cfg.mqtt:
        mqtt = MQTT(cfg.mqtt['ip'], cfg.mqtt.get('port', 1883), cfg.mqtt.get('user', None), cfg.mqtt.get('cred', None))
    resource_args = {
        'command_controller': CommandController(cfg),
        'tivoController': TivoController(cfg, mqtt),
        'pj_controller': PJController(cfg, mqtt),
        'info_provider': info_provider,
        'mqtt': mqtt,
        'config': cfg,
        'version': cfg.version
    }
    app = Flask('cmdserver')
    api = Api(app, prefix='/api', doc='/api/doc/', version=resource_args['version'], title='cmdserver',
              description='Backend api for cmdserver')

    def decorate_ns(ns, p=None):
        for r in ns.resources:
            r.kwargs['resource_class_kwargs'] = resource_args
        api.add_namespace(ns, path=p)

    decorate_ns(commands.api)
    decorate_ns(command.api)
    decorate_ns(playingnow.api)
    decorate_ns(tivos.api)
    decorate_ns(tivo.api)
    decorate_ns(info.api)
    decorate_ns(pj.api)
    decorate_ns(version.api)
    decorate_ns(wake.api)
    return app, ws_server


def main(args=None):
    """ The main routine. """
    cfg = Config('cmdserver')
    app, ws_server = create_app(cfg)
    logger = cfg.configure_logger()

    import logging
    logger = logging.getLogger('twisted')
    from twisted.internet import reactor
    from twisted.web.resource import Resource
    from twisted.web import static, server
    from twisted.web.wsgi import WSGIResource
    from twisted.application import service
    from twisted.internet import endpoints

    class ReactApp:
        """
        Handles the react app (excluding the static dir).
        """

        def __init__(self, path):
            # TODO allow this to load when in debug mode even if the files don't exist
            self.publicFiles = {f: static.File(os.path.join(path, f)) for f in os.listdir(path) if
                                os.path.exists(os.path.join(path, f))} if os.path.exists(path) else {}
            self.indexHtml = ReactIndex(os.path.join(path, 'index.html'))

        def get_file(self, path):
            """
            overrides getChild so it always just serves index.html unless the file does actually exist (i.e. is an
            icon or something like that)
            """
            return self.publicFiles.get(path.decode('utf-8'), self.indexHtml)

    class ReactIndex(static.File):
        """
        a twisted File which overrides getChild so it always just serves index.html (NB: this is a bit of a hack,
        there is probably a more correct way to do this but...)
        """

        def getChild(self, path, request):
            return self

    class FlaskAppWrapper(Resource):
        """
        wraps the flask app as a WSGI resource while allow the react index.html (and its associated static content)
        to be served as the default page.
        """

        def __init__(self):
            super().__init__()
            self.wsgi = WSGIResource(reactor, reactor.getThreadPool(), app)
            import sys
            if getattr(sys, 'frozen', False):
                # pyinstaller lets you copy files to arbitrary locations under the _MEIPASS root dir
                uiRoot = os.path.join(sys._MEIPASS, 'ui')
            elif cfg.webappPath is not None:
                uiRoot = cfg.webappPath
            else:
                # release script moves the ui under the analyser package because setuptools doesn't seem to include
                # files from outside the package
                uiRoot = os.path.join(os.path.dirname(__file__), 'ui')
            if os.path.exists(uiRoot):
                logger.info(f'Serving ui from {uiRoot}')
            else:
                logger.info('No UI, API only')
            self.react = ReactApp(uiRoot)
            self.static = static.File(os.path.join(uiRoot, 'static'))
            self.icons = static.File(cfg.iconPath)
            ws_server.factory.startFactory()
            self.ws_resource = WebSocketResource(ws_server.factory)

        def getChild(self, path, request):
            """
            Overrides getChild to allow the request to be routed to the wsgi app (i.e. flask for the rest api
            calls), the static dir (i.e. for the packaged css/js etc), the various concrete files (i.e. the public
            dir from react-app), the command icons or to index.html (i.e. the react app) for everything else.
            :param path:
            :param request:
            :return:
            """
            # allow CORS (CROSS-ORIGIN RESOURCE SHARING) for debug purposes
            request.setHeader('Access-Control-Allow-Origin', '*')
            request.setHeader('Access-Control-Allow-Methods', 'GET, PUT')
            request.setHeader('Access-Control-Allow-Headers', 'x-prototype-version,x-requested-with')
            request.setHeader('Access-Control-Max-Age', '2520')  # 42 hours
            logger.debug(f"Handling {path}")
            if path == b'ws':
                return self.ws_resource
            elif path == b'api':
                request.prepath.pop()
                request.postpath.insert(0, path)
                return self.wsgi
            elif path == b'static':
                return self.static
            elif path == b'icons':
                return self.icons
            else:
                return self.react.get_file(path)

        def render(self, request):
            return self.wsgi.render(request)

    application = service.Application('cmdserver')
    if cfg.is_access_logging is True:
        site = server.Site(FlaskAppWrapper(), logPath=path.join(cfg.config_path, 'access.log').encode())
    else:
        site = server.Site(FlaskAppWrapper())
    logger.info(f'Listening on 0.0.0.0:{cfg.port}')
    endpoint = endpoints.TCP4ServerEndpoint(reactor, cfg.port, interface='0.0.0.0')
    endpoint.listen(site)
    reactor.run()


if __name__ == '__main__':
    main()



import logging

from flask_restx import Resource, Namespace

logger = logging.getLogger('command')

api = Namespace('1/commands', description='States all configured commands')


@api.route('')
class Commands(Resource):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__controller = kwargs['command_controller']

    def get(self):
        return {'commands': self.__controller.commands}



import logging

from flask import request
from flask_restx import Resource, Namespace

logger = logging.getLogger('tivos')

api = Namespace('1/tivos', description='Access to the tivo api')


@api.route('')
class Tivos(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__controller = kwargs['tivoController']

    def get(self):
        return self.__controller.tivos

    def put(self):
        payload = request.get_json()
        type = payload['type']
        name = payload['name']
        command = payload['command']
        logger.debug('Sending ' + type + '/' + command + ' to ' + name)
        if self.__controller.has_tivo(name):
            try:
                sent = self.__controller.send(name, type, command)
                # sent: [] of commands, channel: current channel text, type, name, command
                return {**sent, **payload}, 200
            except Exception as e:
                return {'error': str(e), **payload}, 500
        else:
            return 404



import json
import logging
import time
from enum import Enum
from threading import Lock
from time import sleep
from typing import Optional, Tuple, Union

from cmdserver.debounce import debounce
from cmdserver.jvc import CommandExecutor, CommandNack
from cmdserver.jvccommands import Command, load_all_commands, Numeric, PictureMode, Anamorphic, PowerState, \
    InstallationMode, READ_ONLY_RC
from cmdserver.mqtt import MQTT

logger = logging.getLogger('pjcontroller')


class PJController:

    def __init__(self, config, mqtt: Optional[MQTT]):
        self.__pj_macros = config.pj_macros
        self.__mqtt = mqtt
        self.__executor = CommandExecutor(host=config.pj_ip) if config.pj_ip else None
        self.__commands = load_all_commands()
        self.__lock = Lock()
        self.__attributes = {
            'anamorphicMode': '',
            'installationMode': '',
            'pictureMode': ''
        }
        self.__looping = None
        if self.__mqtt:
            from twisted.internet import task
            self.__looping = task.LoopingCall(self.__update_state)
            logger.info(f'Initiating PJ refresh every 20s')
            self.__looping.start(20, now=False)

    def __update_state(self):
        update_in = 0.5
        with self.__lock:
            logger.info('Refreshing PJ State')
            cmd = Command.Power
            try:
                self.__connect()
                self.__mqtt.online('pj')
                power = self.__executor.get(cmd)
                if power == PowerState.LampOn:
                    cmd = Command.Anamorphic
                    ana = self.__executor.get(cmd)
                    cmd = Command.PictureMode
                    pic = self.__executor.get(cmd)
                    cmd = Command.InstallationMode
                    install = self.__executor.get(cmd)
                    self.__attributes = {
                        'anamorphicMode': ana.name,
                        'installationMode': install.name,
                        'pictureMode': pic.name
                    }
                    update_in = 10
                else:
                    update_in = 1 if power == PowerState.Starting or power == PowerState.Cooling else 20
                self.__mqtt.publish('pj/state', power.name)
                self.__mqtt.publish('pj/attributes', json.dumps(self.__attributes))
                self.__update_state_in(update_in=update_in)
                self.__disconnect()
                logger.info('Refreshed PJ State')
            except CommandNack:
                self.__update_state_in(update_in=update_in)
                self.__disconnect()
                logger.exception(f"Command NACKed - GET {cmd}")
            except:
                self.__mqtt.offline('pj')
                self.__update_state_in(update_in=update_in)
                self.__disconnect()
                logger.exception(f"Unexpected failure while executing cmd: {cmd}")

    def __update_state_in(self, update_in: float = 20, reason: str = None):
        suffix = f' due to {reason}' if reason else ''
        if self.__looping.interval != update_in:
            logger.info(f'Changing PJ refresh rate from {self.__looping.interval} to {update_in}s{suffix}')
            self.__looping.stop()
            self.__looping.start(update_in, now=False)

    @property
    def state(self):
        return self.__attributes

    @property
    def enabled(self):
        return self.__executor is not None

    def __connect(self):
        self.__executor.connect()

    @debounce(4)
    def __disconnect(self):
        self.__executor.disconnect(fail=False)

    def get(self, command):
        with self.__lock:
            try:
                cmd = Command[command]
                self.__connect()
                val = self.__executor.get(cmd)
                self.__disconnect()
                if val is not None:
                    return val.name if isinstance(val, Enum) else val.replace('"', '').strip()
                return val
            except KeyError:
                logger.warning(f"Ignoring unknown command {command}")
                return None
            except CommandNack:
                logger.exception(f"Command NACKed - GET {command}")
                return -1
            except:
                logger.exception(f"Unexpected failure while executing cmd: {command}")
                return -1

    def send(self, commands):
        """ Sends the commands to the PJ """
        with self.__lock:
            sent = []
            vals = []
            self.__connect()
            for command in commands:
                if command[0:5] == 'PAUSE':
                    sleep_secs = float(command[5:])
                    logger.info(f"Sleeping for {sleep_secs:.3f}")
                    sleep(sleep_secs)
                else:
                    if command in self.__pj_macros:
                        for cmd in self.__pj_macros[command]:
                            c, e, v = self.__execute(cmd)
                        if c and e:
                            sent.append((c, e))
                        if v:
                            vals.append(v)
                    else:
                        c, e, v = self.__execute(command)
                        if c and e:
                            sent.append((c, e))
                        if v:
                            vals.append(v)
            self.__disconnect()
            if self.__mqtt:
                self.__update_state_if_necessary(sent)
            return vals

    def __update_state_if_necessary(self, sent):
        mutate_cmd = None
        for c, e in sent:
            if c != Command.Remote:
                mutate_cmd = f'{c} - {e}'
                break
            else:
                if e not in READ_ONLY_RC:
                    mutate_cmd = f'{c} - {e}'
                    break
        if mutate_cmd:
            self.__update_state_in(update_in=1, reason=f'{mutate_cmd}')
        else:
            logger.info(f'Info only, PJ state will not be updated')

    def __execute(self, cmd) -> Tuple[Optional[Command], Optional[Union[Enum, Numeric]], any]:
        tokens = cmd.split('.')
        if len(tokens) > 1:
            try:
                cmd_enum = Command[tokens[0]]
                if isinstance(cmd_enum.value, tuple):
                    cmd_arg_enum = cmd_enum.value[1]
                    if cmd_arg_enum.__name__ == tokens[1]:
                        logger.info(f"Executing {cmd}")
                        if issubclass(cmd_arg_enum, Enum):
                            tok = cmd_arg_enum[tokens[2]]
                            return cmd_enum, tok, self.__executor.set(cmd_enum, tok)
                        elif issubclass(cmd_arg_enum, Numeric):
                            tok = Numeric(int(tokens[2]))
                            return cmd_enum, tok, self.__executor.set(cmd_enum, tok)
                        else:
                            logger.warning(f"Unsupported value type for {cmd} - {cmd_arg_enum.__name__}")
            except (AttributeError, KeyError):
                logger.exception(f"Ignoring unknown command {cmd}")
            except:
                logger.exception(f"Unexpected exception while processing {cmd}")
        else:
            logger.error(f"Ignoring unknown command {cmd}")
        return None, None, None



from threading import Timer

# from https://gist.github.com/walkermatt/2871026


def debounce(wait):
    """ Decorator that will postpone a functions
        execution until after wait seconds
        have elapsed since the last time it was invoked. """
    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)
            try:
                debounced.t.cancel()
            except AttributeError:
                pass
            debounced.t = Timer(wait, call_it)
            debounced.t.start()
        return debounced
    return decorator
